from __future__ import division

import numpy as np
import os
import qubic
from pysimulators import FitsArray
from qubic import (
    QubicCalibration, QubicAcquisition, QubicInstrument, QubicSampling)

MAPPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                       'syn256_pol.fits')
DATAPATH = os.path.join(os.path.dirname(__file__), 'data')
map_orig = qubic.io.read_map(MAPPATH, field='I_STOKES')


def test_ga():
    p = FitsArray(os.path.join(DATAPATH, 'ptg_np100_10deg.fits'))
    o = QubicSampling(azimuth=p[..., 0], elevation=p[..., 1],
                      pitch=p[..., 2])
    t = FitsArray(os.path.join(DATAPATH, 'tod_ndet16_np100_ga.fits'))
    c = QubicCalibration(os.path.join(DATAPATH, 'calfiles'))
    i = QubicInstrument('monochromatic,nopol', c, ngrids=1, nu=150e9, dnu_nu=0,
                        nside=256, synthbeam_fraction=0.99)
    acq = QubicAcquisition(i, o)
    C = acq.get_convolution_peak_operator(fwhm=np.radians(0.64883707))
    P = acq.get_projection_operator()
    H = P * C
    t2 = H(map_orig)
    np.testing.assert_almost_equal(t, t2)



from __future__ import division
import numpy as np
from pyoperators import MPI
from numpy.testing import assert_equal
from pyoperators.utils.testing import assert_same
from qubic import QubicAcquisition, QubicInstrument, create_random_pointings
from qubic.mapmaking import tod2map_all, tod2map_each

rank = MPI.COMM_WORLD.rank
size = MPI.COMM_WORLD.size


def test():
    kinds = 'I', 'IQU'
    instrument = QubicInstrument(synthbeam_dtype=float)[:400]
    np.random.seed(0)
    sampling = create_random_pointings([0, 90], 30, 5)
    sampling.angle_hwp = np.random.random_integers(0, 7, len(sampling)) * 11.25
    skies = np.ones(12 * 256**2), np.ones((12 * 256**2, 3))

    def func(sampling, kind, sky, ref1, ref2, ref3, ref4, ref5, ref6):
        nprocs_instrument = max(size // 2, 1)
        acq = QubicAcquisition(instrument, sampling, kind=kind,
                               nprocs_instrument=nprocs_instrument)
        assert_equal(acq.comm.size, size)
        assert_equal(acq.instrument.detector.comm.size, nprocs_instrument)
        assert_equal(acq.sampling.comm.size, size / nprocs_instrument)
        H = acq.get_operator()
        invntt = acq.get_invntt_operator()
        tod = H(sky)
        #actual1 = acq.unpack(H(sky))
        #assert_same(actual1, ref1, atol=10)
        actual2 = H.T(invntt(tod))
        assert_same(actual2, ref2, atol=10)
        actual2 = (H.T * invntt * H)(sky)
        assert_same(actual2, ref2, atol=10)
        actual3, actual4 = tod2map_all(acq, tod, disp=False, maxiter=2)
        assert_same(actual3, ref3, atol=10)
        assert_same(actual4, ref4)
        #actual5, actual6 = tod2map_each(acq, tod, disp=False)
        #assert_same(actual5, ref5, atol=1000)
        #assert_same(actual6, ref6)

    for kind, sky in zip(kinds, skies):
        acq = QubicAcquisition(instrument, sampling, kind=kind,
                               comm=MPI.COMM_SELF)
        assert_equal(acq.comm.size, 1)
        H = acq.get_operator()
        invntt = acq.get_invntt_operator()
        tod = H(sky)
        ref1 = acq.unpack(tod)
        ref2 = H.T(invntt(tod))
        ref3, ref4 = tod2map_all(acq, tod, disp=False, maxiter=2)
        ref5, ref6 = None, None #tod2map_each(acq, tod, disp=False)
        yield (func, sampling, kind, sky, ref1, ref2, ref3, ref4, ref5, ref6)



from __future__ import division

import healpy as hp
import matplotlib.pyplot as mp
import numpy as np
import os
import qubic
from qubic import (
    QubicAcquisition, create_sweeping_pointings, equ2gal, map2tod, tod2map_all,
    tod2map_each)

DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
x0 = qubic.io.read_map(DATAPATH)

# parameters
nside = 256
racenter = 0.0      # deg
deccenter = -57.0   # deg
angspeed = 1        # deg/sec
delta_az = 15.      # deg
angspeed_psi = 0.1  # deg/sec
maxpsi = 45.        # deg
nsweeps_el = 300
duration = 24       # hours
ts = 20             # seconds

# get the sampling model
np.random.seed(0)
sampling = create_sweeping_pointings(
    [racenter, deccenter], duration, ts, angspeed, delta_az, nsweeps_el,
    angspeed_psi, maxpsi)
sampling.angle_hwp = np.random.random_integers(0, 7, len(sampling)) * 11.25

# get the acquisition model
acquisition = QubicAcquisition(150, sampling,
                               nside=nside,
                               synthbeam_fraction=0.99,
                               detector_tau=0.01,
                               detector_sigma=1.,
                               detector_fknee=1.,
                               detector_fslope=1)

# simulate the timeline
tod, x0_convolved = map2tod(acquisition, x0, convolution=True)
tod_noisy = tod + acquisition.get_noise()

# reconstruct using two methods
map_all, cov_all = tod2map_all(acquisition, tod, tol=1e-2)
map_each, cov_each = tod2map_each(acquisition, tod, tol=1e-2)


# some display
def display(map, cov, msg, sub):
    for i, (kind, lim) in enumerate(zip('IQU', [200, 10, 10])):
        map_ = map[..., i].copy()
        mask = cov == 0
        map_[mask] = np.nan
        hp.gnomview(map_, rot=center, reso=5, xsize=400, min=-lim, max=lim,
                    title=msg + ' ' + kind, sub=(3, 3, 3 * (sub-1) + i+1))

center = equ2gal(racenter, deccenter)

mp.figure(1)
display(x0_convolved, cov_all, 'Original map', 1)
display(map_all, cov_all, 'Reconstructed map', 2)
display(map_all - x0_convolved, cov_all, 'Difference map', 3)
mp.show()

mp.figure(2)
display(x0_convolved, cov_each, 'Original map', 1)
display(map_each, cov_each, 'Reconstructed map', 2)
display(map_each - x0_convolved, cov_each, 'Difference map', 3)
mp.show()



# coding: utf-8
from __future__ import division

try:
    import matplotlib.pyplot as mp
except:
    pass
import numexpr as ne
import numpy as np
from pyoperators import (
    DenseBlockDiagonalOperator, IdentityOperator, HomothetyOperator,
    ReshapeOperator, Rotation2dOperator, Rotation3dOperator,
    Spherical2CartesianOperator)
from pyoperators.utils import pool_threading
from pysimulators import (
    Instrument, LayoutVertex, ConvolutionTruncatedExponentialOperator,
    ProjectionOperator)
from pysimulators.interfaces.healpy import Cartesian2HealpixOperator
from pysimulators.sparse import (
    FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix)
from scipy.constants import c, pi
from . import _flib as flib
from .beams import GaussianBeam
from .calibration import QubicCalibration
from .utils import _compress_mask

__all__ = ['QubicInstrument',
           'SimpleInstrument']


class SimpleInstrument(Instrument):
    """
    The SimpleInstrument class. Classical imager with a well-behaved
    single-peak beam.

    """
    def __init__(self, calibration=None, detector_fknee=0, detector_fslope=1,
                 detector_ncorr=10, detector_ngrids=2, detector_sigma=10,
                 detector_tau=0.01, synthbeam_dtype=np.float32, **keywords):
        """
        Parameters
        ----------
        calibration : QubicCalibration
            The calibration tree.
        detector_fknee : array-like
            The detector 1/f knee frequency in Hertz.
        detector_fslope : array-like
            The detector 1/f slope index.
        detector_ncorr : int
            The detector 1/f correlation length.
        detector_ngrids : int, optional
            Number of detector grids.
        detector_sigma : array-like
            The standard deviation of the detector white noise component.
        detector_tau : array-like
            The detector time constants in seconds.
        synthbeam_dtype : dtype, optional
            The data type for the synthetic beams (default: float32).
            It is the dtype used to store the values of the pointing matrix.

        """
        if calibration is None:
            calibration = QubicCalibration()
        self.calibration = calibration
        layout = self._get_detector_layout(
            detector_ngrids, detector_sigma, detector_fknee, detector_fslope,
            detector_ncorr, detector_tau)
        Instrument.__init__(self, 'QUBIC', layout)
        self._init_optics(**keywords)
        self._init_synthetic_beam(synthbeam_dtype)

    def _get_detector_layout(self, ngrids, sigma, fknee, fslope, ncorr,
                             tau):
        shape, vertex, removed_, index, quadrant = \
            self.calibration.get('detarray')
        if ngrids == 2:
            shape = (2,) + shape
            vertex = np.array([vertex, vertex])
            removed_ = np.array([removed_, removed_])
            index = np.array([index, index + np.max(index) + 1], index.dtype)
            quadrant = np.array([quadrant, quadrant + 4], quadrant.dtype)
        layout = LayoutVertex(
            shape, 4, vertex=vertex, selection=~removed_, ordering=index,
            quadrant=quadrant, sigma=sigma, fknee=fknee, fslope=fslope,
            tau=tau)
        layout.ncorr = ncorr
        layout.ngrids = ngrids
        return layout

    def _init_optics(self, **keywords):
        class Optics(object):
            pass
        optics = Optics()
        optics.focal_length = self.calibration.get('optics')['focal length']
        self.optics = optics

    def _init_synthetic_beam(self, dtype):
        class SyntheticBeam(object):
            pass
        sb = SyntheticBeam()
        sb.dtype = np.dtype(dtype)
        self.synthetic_beam = sb

    def __str__(self):
        state = [('ngrids', self.detector.ngrids),
                 ('selection', _compress_mask(~self.detector.all.removed))]
        return 'Instrument:\n' + \
               '\n'.join(['    ' + a + ': ' + repr(v) for a, v in state]) + \
               '\n\nCalibration:\n' + '\n'. \
               join('    ' + l for l in str(self.calibration).splitlines())

    __repr__ = __str__

    def plot(self, autoscale=True, **keywords):
        """
        Plot detectors on the image plane.

        """
        a = mp.gca()
        vertex = self.detector.vertex
        for v in vertex:
            a.add_patch(mp.Polygon(v, closed=True, fill=False, **keywords))
        if autoscale:
            mp.autoscale()
        mp.show()

    def get_noise(self, sampling, out=None):
        """
        Return a noisy timeline.

        """
        return Instrument.get_noise(
            self, sampling, sigma=self.detector.sigma,
            fknee=self.detector.fknee, fslope=self.detector.fslope, out=out)

    def get_detector_response_operator(self, sampling, tau=None):
        """
        Return the operator for the bolometer responses.

        """
        if tau is None:
            tau = self.detector.tau
        sampling_period = sampling.period
        shapein = len(self), len(sampling)
        if sampling_period == 0:
            return IdentityOperator(shapein)
        return ConvolutionTruncatedExponentialOperator(
            tau / sampling_period, shapein=shapein)

    def get_hwp_operator(self, sampling, scene):
        """
        Return the rotation matrix for the half-wave plate.

        """
        shape = (len(self), len(sampling))
        if scene.kind == 'I':
            return IdentityOperator(shapein=shape)
        if scene.kind == 'QU':
            return Rotation2dOperator(-4 * sampling.angle_hwp,
                                      degrees=True, shapein=shape + (2,))
        return Rotation3dOperator('X', -4 * sampling.angle_hwp,
                                  degrees=True, shapein=shape + (3,))

    def get_invntt_operator(self, sampling):
        """
        Return the inverse time-time noise correlation matrix as an Operator.

        """
        return Instrument.get_invntt_operator(
            self, sampling, sigma=self.detector.sigma,
            fknee=self.detector.fknee, fslope=self.detector.fslope,
            ncorr=self.detector.ncorr)

    def get_polarizer_operator(self, sampling, scene):
        """
        Return operator for the polarizer grid.

        """
        if scene.kind == 'I':
            return HomothetyOperator(1 / self.detector.ngrids)

        if self.detector.ngrids == 1:
            raise ValueError(
                'Polarized input not handled by a single detector grid.')

        nd = len(self)
        nt = len(sampling)
        grid = self.detector.quadrant // 4
        z = np.zeros(nd)
        data = np.array([z + 0.5, 0.5 - grid, z]).T[:, None, None, :]
        return ReshapeOperator((nd, nt, 1), (nd, nt)) * \
            DenseBlockDiagonalOperator(data, shapein=(nd, nt, 3))

    def get_projection_operator(self, sampling, scene, verbose=True):
        """
        Return the peak sampling operator.

        Parameters
        ----------
        sampling : QubicSampling
            The pointing information.
        scene : QubicScene
            The observed scene.
        verbose : bool, optional
            If true, display information about the memory allocation.

        """
        rotation = sampling.cartesian_galactic2instrument
        dtype = self.synthetic_beam.dtype
        ndetectors = len(self)
        ntimes = len(sampling)
        nside = scene.nside

        theta, phi, vals = self._peak_angles(scene)
        ncolmax = theta.shape[-1]
        thetaphi = _pack_vector(theta, phi)  # (ndetectors, ncolmax, 2)
        direction = Spherical2CartesianOperator('zenith,azimuth')(thetaphi)
        e_nf = direction[:, None, :, :]
        if nside > 8192:
            dtype_index = np.dtype(np.int64)
        else:
            dtype_index = np.dtype(np.int32)

        cls = {'I': FSRMatrix,
               'QU': FSRRotation2dMatrix,
               'IQU': FSRRotation3dMatrix}[scene.kind]
        ndims = len(scene.kind)
        s = cls((ndetectors * ntimes * ndims, 12 * nside**2 * ndims),
                ncolmax=ncolmax, dtype=dtype, dtype_index=dtype_index,
                verbose=verbose)

        index = s.data.index.reshape((ndetectors, ntimes, ncolmax))

        def func_thread(i):
            # e_nf[i] shape: (1, ncolmax, 3)
            # e_ni shape: (ntimes, ncolmax, 3)
            e_ni = rotation.T(e_nf[i].swapaxes(0, 1)).swapaxes(0, 1)
            index[i] = Cartesian2HealpixOperator(nside)(e_ni)

        with pool_threading() as pool:
            pool.map(func_thread, xrange(ndetectors))

        if scene.kind == 'I':
            value = s.data.value.reshape(ndetectors, ntimes, ncolmax)
            value[...] = vals[:, None, :]
            shapeout = (ndetectors, ntimes)
        else:
            func = 'pointing_matrix_rot{0}d_i{1}_m{2}'.format(
                ndims, dtype_index.itemsize, dtype.itemsize)
            try:
                getattr(flib.polarization, func)(
                    rotation.data.T, direction.T, s.data.ravel().view(np.int8),
                    vals.T)
            except AttributeError:
                raise TypeError(
                    'The projection matrix cannot be created with types: {0} a'
                    'nd {1}.'.format(dtype, dtype_index))
            if scene.kind == 'QU':
                shapeout = (ndetectors, ntimes, 2)
            else:
                shapeout = (ndetectors, ntimes, 3)

        return ProjectionOperator(s, shapeout=shapeout)

    def _peak_angles(self, scene):
        """
        Return the spherical coordinates (theta, phi) of the beam peaks,
        in radians.

        """
        local_dict = {'f': self.optics.focal_length,
                      'x': np.ascontiguousarray(self.detector.center[..., 0]),
                      'y': np.ascontiguousarray(self.detector.center[..., 1])}
        theta = ne.evaluate('arctan2(sqrt(x**2 + x**2), f)',
                            local_dict=local_dict).reshape(-1, 1)
        phi = ne.evaluate('arctan2(y, x)+pi',
                          local_dict=local_dict).reshape(-1, 1)
        vals = np.ones_like(theta)
        return theta, phi, vals


class QubicInstrument(SimpleInstrument):
    """
    The QubicInstrument class. It represents the instrument setup.

    """
    def __init__(self, calibration=None, detector_fknee=0, detector_fslope=1,
                 detector_ncorr=10, detector_ngrids=2, detector_sigma=10,
                 detector_tau=0.01, primary_beam=None, secondary_beam=None,
                 synthbeam_dtype=np.float32, synthbeam_fraction=0.99,
                 **keywords):
        """
        Parameters
        ----------
        calibration : QubicCalibration
            The calibration tree.
        detector_fknee : array-like
            The detector 1/f knee frequency in Hertz.
        detector_fslope : array-like
            The detector 1/f slope index.
        detector_ncorr : int
            The detector 1/f correlation length.
        detector_ngrids : int, optional
            Number of detector grids.
        detector_sigma : array-like
            The standard deviation of the detector white noise component.
        detector_tau : array-like
            The detector time constants in seconds.
        primary_beam : function f(theta [rad], phi [rad])
            The primary beam transmission function.
        secondary_beam : function f(theta [rad], phi [rad])
            The secondary beam transmission function.
        synthbeam_dtype : dtype, optional
            The data type for the synthetic beams (default: float32).
            It is the dtype used to store the values of the pointing matrix.
        synthbeam_fraction: float, optional
            The fraction of significant peaks retained for the computation
            of the synthetic beam.

        """
        if calibration is None:
            calibration = QubicCalibration()
        SimpleInstrument.__init__(
            self, calibration, detector_fknee, detector_fslope, detector_ncorr,
            detector_ngrids, detector_sigma, detector_tau, synthbeam_dtype,
            **keywords)
        self._init_beams(primary_beam, secondary_beam)
        self._init_horns()
        self.synthetic_beam.fraction = synthbeam_fraction
        self.synthetic_beam.kmax = 8  # all peaks are considered

    def _init_beams(self, primary, secondary):
        if primary is None:
            primary = GaussianBeam(self.calibration.get('primbeam'))
        self.primary_beam = primary
        if secondary is None:
            secondary = GaussianBeam(self.calibration.get('primbeam'),
                                     backward=True)
        self.secondary_beam = secondary

    def _init_horns(self):
        self.horn = self.calibration.get('hornarray')

    def __str__(self):
        state = [('synthbeam_fraction', self.synthetic_beam.fraction)]
        out = SimpleInstrument.__str__(self)
        index = out.index('\nCalibration')
        return out[:index] + \
               '\n'.join(['    {0}: {1!r}'.format(*_) for _ in state]) + \
               out[index:]

    __repr__ = __str__

    def _peak_angles(self, scene):
        fraction = self.synthetic_beam.fraction
        theta, phi = self._peak_angles_kmax(scene, self.synthetic_beam.kmax)
        val = np.array(self.primary_beam(theta, phi), dtype=float, copy=False)
        val[~np.isfinite(val)] = 0
        val /= np.sum(val, axis=-1)[:, None]
        index = _argsort_reverse(val)
        theta = theta[index]
        phi = phi[index]
        val = val[index]
        cumval = np.cumsum(val, axis=-1)
        imaxs = np.argmax(cumval >= fraction, axis=-1) + 1
        imax = max(imaxs)

        # slice initial arrays to discard the non-significant peaks
        theta = theta[:, :imax]
        phi = phi[:, :imax]
        val = val[:, :imax]

        # remove additional per-detector non-significant peaks
        # and remove potential NaN in theta, phi
        for idet, imax_ in enumerate(imaxs):
            val[idet, imax_:] = 0
            theta[idet, imax_:] = pi / 2 #XXX 0 leads to NaN
            phi[idet, imax_:] = 0
        val /= np.sum(val, axis=-1)[:, None]
        return theta, phi, val

    def _peak_angles_kmax(self, scene, kmax):
        """
        Return the spherical coordinates (theta, phi) of the beam peaks,
        in radians.

        Parameters
        ----------
        kmax : int, optional
            The diffraction order above which the peaks are ignored.
            For instance, a value of kmax=2 will model the synthetic beam by
            (2 * kmax + 1)**2 = 25 peaks and a value of kmax=0 will only sample
            the central peak.
        """
        ndetector = len(self)
        center = self.detector.center
        lmbda = c / scene.nu
        dx = self.horn.spacing
        detvec = np.vstack([-center[..., 0],
                            -center[..., 1],
                            np.zeros(ndetector) + self.optics.focal_length]).T
        detvec.T[...] /= np.sqrt(np.sum(detvec**2, axis=1))

        kx, ky = np.mgrid[-kmax:kmax+1, -kmax:kmax+1]
        nx = detvec[:, 0, np.newaxis] - lmbda * kx.ravel() / dx
        ny = detvec[:, 1, np.newaxis] - lmbda * ky.ravel() / dx
        local_dict = {'nx': nx, 'ny': ny}
        theta = ne.evaluate('arcsin(sqrt(nx**2 + ny**2))',
                            local_dict=local_dict)
        phi = ne.evaluate('arctan2(ny, nx)', local_dict=local_dict)
        return theta, phi


def _argsort_reverse(a, axis=-1):
    i = list(np.ogrid[[slice(x) for x in a.shape]])
    i[axis] = a.argsort(axis)[:, ::-1]
    return i


def _pack_vector(*args):
    shape = np.broadcast(*args).shape
    out = np.empty(shape + (len(args),))
    for i, arg in enumerate(args):
        out[..., i] = arg
    return out



# coding: utf-8
from __future__ import division

import numpy as np
from astropy.time import Time, TimeDelta
from numpy.random import random_sample as randomu
from pyoperators import (
    Cartesian2SphericalOperator, Rotation3dOperator,
    Spherical2CartesianOperator, rule_manager)
from pyoperators.utils import isscalarlike
from pysimulators import (
    PointingHorizontal, CartesianEquatorial2GalacticOperator,
    CartesianEquatorial2HorizontalOperator,
    CartesianHorizontal2EquatorialOperator,
    CartesianGalactic2EquatorialOperator,
    SphericalEquatorial2GalacticOperator,
    SphericalGalactic2EquatorialOperator,
    SphericalEquatorial2HorizontalOperator,
    SphericalHorizontal2EquatorialOperator)
from pysimulators.interfaces.healpy import Cartesian2HealpixOperator

__all__ = ['QubicSampling',
           'create_random_pointings',
           'create_sweeping_pointings',
           'equ2gal',
           'equ2hor',
           'gal2equ',
           'gal2hor',
           'hor2equ',
           'hor2gal']

DOMECLAT = -(75 + 6 / 60)
DOMECLON = 123 + 20 / 60


class QubicSampling(PointingHorizontal):
    """
    Attributes
    ----------
    azimuth : array-like
        The pointing azimuth [degrees].
    elevation : array-like
        The pointing elevation [degrees].
    pitch : array-like
        The instrument pitch angle [degrees].
    angle_hwp : array-like
        The half-wave plate angle [degrees].
    time : array-like
        Elapsed time for each pointing since the observation start date,
        in seconds.
    date_obs : string or astropy.time.Time
        The observation start date.
    period : float
        The sampling period [s].
    latitude : float
        Telescope latitude [degrees].
    longitude : float
        Telescope longitude [degrees].

    Examples
    --------
    pointing = QubicSampling(azimuth=azimuth, elevation=elevation,
                             angle_hwp=angle_hwp, period=...)
    pointing = QubicSampling(azimuth, elevation, pitch, angle_hwp, period=...)

    """
    DEFAULT_DATE_OBS = '2016-01-01 00:00:00'
    DEFAULT_PERIOD = 1
    DEFAULT_LATITUDE = DOMECLAT
    DEFAULT_LONGITUDE = DOMECLON

    def __init__(self, *args, **keywords):
        angle_hwp = keywords.get('angle_hwp', 0)
        if len(args) == 4:
            args = list(args)
            angle_hwp = args.pop()
        PointingHorizontal.__init__(self, angle_hwp=angle_hwp, healpix=None,
                                    *args, **keywords)

    def healpix(self, nside):
        time = self.date_obs + TimeDelta(self.time, format='sec')
        c2h = Cartesian2HealpixOperator(nside)
        e2g = CartesianEquatorial2GalacticOperator()
        h2e = CartesianHorizontal2EquatorialOperator(
            'NE', time, self.latitude, self.longitude)
        rotation = c2h(e2g(h2e))
        return rotation(self.cartesian)

    @property
    def cartesian_galactic2instrument(self):
        """
        Return the galactic-to-instrument transform.

        """
        time = self.date_obs + TimeDelta(self.time, format='sec')
        with rule_manager(none=False):
            r = Rotation3dOperator("ZY'Z''", self.azimuth, 90 - self.elevation,
                                   self.pitch, degrees=True).T * \
                CartesianEquatorial2HorizontalOperator(
                    'NE', time, self.latitude, self.longitude) * \
                CartesianGalactic2EquatorialOperator()
        return r

    @property
    def cartesian_instrument2galactic(self):
        return self.cartesian_galactic2instrument.I


QubicPointing = QubicSampling


def create_random_pointings(center, npointings, dtheta, date_obs=None,
                            period=None, latitude=None, longitude=None):
    """
    Return pointings randomly and uniformly distributed in a spherical cap.

    Parameters
    ----------
    center : 2-tuple
        The R.A. and declination of the center of the FOV, in degrees.
    npointings : int
        The number of requested pointings
    dtheta : float
        The maximum angular distance to the center.
    date_obs : str or astropy.time.Time, optional
        The starting date of the observation (UTC).
    period : float, optional
        The sampling period of the pointings, in seconds.
    latitude : float, optional
        The observer's latitude [degrees]. Default is DOMEC's.
    longitude : float, optional
        The observer's longitude [degrees]. Default is DOMEC's.

    """
    cosdtheta = np.cos(np.radians(dtheta))
    theta = np.degrees(np.arccos(cosdtheta +
                                 (1 - cosdtheta) * randomu(npointings)))
    phi = randomu(npointings) * 360
    pitch = randomu(npointings) * 360
    p = QubicSampling(
        npointings, date_obs=date_obs, period=period, latitude=latitude,
        longitude=longitude)
    time = p.date_obs + TimeDelta(p.time, format='sec')
    c2s = Cartesian2SphericalOperator('azimuth,elevation', degrees=True)
    e2h = CartesianEquatorial2HorizontalOperator(
        'NE', time, p.latitude, p.longitude)
    rot = Rotation3dOperator("ZY'", center[0], 90-center[1], degrees=True)
    s2c = Spherical2CartesianOperator('zenith,azimuth', degrees=True)
    rotation = c2s(e2h(rot(s2c)))
    coords = rotation(np.asarray([theta.T, phi.T]).T)
    p.azimuth = coords[..., 0]
    p.elevation = coords[..., 1]
    p.pitch = pitch
    return p


def create_sweeping_pointings(
        center, duration, period, angspeed, delta_az, nsweeps_per_elevation,
        angspeed_psi, maxpsi, date_obs=None, latitude=None, longitude=None):
    """
    Return pointings according to the sweeping strategy:
    Sweep around the tracked FOV center azimuth at a fixed elevation, and
    update elevation towards the FOV center at discrete times.

    Parameters
    ----------
    center : array-like of size 2
        The R.A. and Declination of the center of the FOV.
    duration : float
        The duration of the observation, in hours.
    period : float
        The sampling period of the pointings, in seconds.
    angspeed : float
        The pointing angular speed, in deg / s.
    delta_az : float
        The sweeping extent in degrees.
    nsweeps_per_elevation : int
        The number of sweeps during a phase of constant elevation.
    angspeed_psi : float
        The pitch angular speed, in deg / s.
    maxpsi : float
        The maximum pitch angle, in degrees.
    latitude : float, optional
        The observer's latitude [degrees]. Default is DOMEC's.
    longitude : float, optional
        The observer's longitude [degrees]. Default is DOMEC's.
    date_obs : str or astropy.time.Time, optional
        The starting date of the observation (UTC).
    return_hor : bool, optional
        Obsolete keyword.

    Returns
    -------
    pointings : QubicPointing
        Structured array containing the azimuth, elevation and pitch angles,
        in degrees.

    """
    nsamples = int(np.ceil(duration * 3600 / period))
    out = QubicPointing(
        nsamples, date_obs=date_obs, period=period, latitude=latitude,
        longitude=longitude)
    racenter = center[0]
    deccenter = center[1]
    backforthdt = delta_az / angspeed * 2

    # compute the sweep number
    isweeps = np.floor(out.time / backforthdt).astype(int)

    # azimuth/elevation of the center of the field as a function of time
    azcenter, elcenter = equ2hor(racenter, deccenter, out.time,
                                 date_obs=out.date_obs, latitude=out.latitude,
                                 longitude=out.longitude)

    # compute azimuth offset for all time samples
    daz = out.time * angspeed
    daz = daz % (delta_az * 2)
    mask = daz > delta_az
    daz[mask] = -daz[mask] + 2 * delta_az
    daz -= delta_az / 2

    # elevation is kept constant during nsweeps_per_elevation
    elcst = np.zeros(nsamples)
    ielevations = isweeps // nsweeps_per_elevation
    nelevations = ielevations[-1] + 1
    for i in xrange(nelevations):
        mask = ielevations == i
        elcst[mask] = np.mean(elcenter[mask])

    # azimuth and elevations to use for pointing
    azptg = azcenter + daz
    elptg = elcst

    ### scan psi as well
    pitch = out.time * angspeed_psi
    pitch = pitch % (4 * maxpsi)
    mask = pitch > (2 * maxpsi)
    pitch[mask] = -pitch[mask] + 4 * maxpsi
    pitch -= maxpsi

    out.azimuth = azptg
    out.elevation = elptg
    out.pitch = pitch
    return out


def _format_sphconv(a, b, date_obs=None, time=None):
    incoords = np.empty(np.broadcast(a, b).shape + (2,))
    incoords[..., 0] = a
    incoords[..., 1] = b
    if date_obs is None:
        return incoords
    time = Time(date_obs if isscalarlike(time)
                         else [date_obs], scale='utc') + \
           TimeDelta(time, format='sec')
    return incoords, time


def equ2gal(ra, dec):
    """
    equ2gal(ra, dec) -> l, b
    Equatorial to galactic spherical conversion. Angles are in degrees.

    """
    incoords = _format_sphconv(ra, dec)
    outcoords = SphericalEquatorial2GalacticOperator(degrees=True)(incoords)
    return outcoords[..., 0], outcoords[..., 1]


def gal2equ(l, b):
    """
    gal2equ(l, b) -> ra, dec
    Galactic to equatorial spherical conversion. Angles are in degrees.

    """
    incoords = _format_sphconv(l, b)
    outcoords = SphericalGalactic2EquatorialOperator(degrees=True)(incoords)
    return outcoords[..., 0], outcoords[..., 1]


def equ2hor(ra, dec, time, date_obs=QubicPointing.DEFAULT_DATE_OBS,
            latitude=DOMECLAT, longitude=DOMECLON):
    """
    equ2hor(ra, dec, time, [date_obs, [latitude, [longitude]]]) -> az, el
    Equatorial to horizontal spherical conversion. Angles are in degrees.

    Parameters
    ----------
    time : array-like
        Elapsed time in seconds since date_obs.
    date_obs : string
        The starting date, UTC.
    latitude : float
        The observer's latitude geolocation. Default is Dome C.
    longitude : float
        The observer's longitude geolocation. Default is Dome C.

    Example
    -------
    >>> equ2hor(0, 0, 0, date_obs='2000-01-01 00:00:00')
    (array(135.71997181016644), array(-10.785386358099927))

    """
    incoords, time = _format_sphconv(ra, dec, date_obs, time)
    outcoords = SphericalEquatorial2HorizontalOperator(
        'NE', time, latitude, longitude, degrees=True)(incoords)
    return outcoords[..., 0], outcoords[..., 1]


def hor2equ(azimuth, elevation, time, date_obs=QubicPointing.DEFAULT_DATE_OBS,
            latitude=DOMECLAT, longitude=DOMECLON):
    """
    hor2equ(az, el, time, [date_obs, [latitude, [longitude]]]) -> ra, dec
    Horizontal to equatorial spherical conversion. Angles are in degrees.


    Parameters
    ----------
    time : array-like
        Elapsed time in seconds since date_obs.
    date_obs : string
        The starting date, UTC.
    latitude : float
        The observer's latitude geolocation. Default is Dome C.
    longitude : float
        The observer's longitude geolocation. Default is Dome C.

    Example
    -------
    >>> hor2equ(135.71997181016644, -10.785386358099927, 0,
    ...         date_obs='2000-01-01 00:00:00')
    (array(1.1927080055488187e-14), array(-1.2722218725854067e-14))

    """
    incoords, time = _format_sphconv(azimuth, elevation, date_obs, time)
    outcoords = SphericalHorizontal2EquatorialOperator(
        'NE', time, latitude, longitude, degrees=True)(incoords)
    return outcoords[..., 0], outcoords[..., 1]


def gal2hor(l, b, time, date_obs=QubicPointing.DEFAULT_DATE_OBS,
            latitude=DOMECLAT, longitude=DOMECLON):
    """
    gal2hor(l, b, time, [date_obs, [latitude, [longitude]]]) -> az, el
    Galactic to horizontal spherical conversion. Angles are in degrees.


    Parameters
    ----------
    time : array-like
        Elapsed time in seconds since date_obs.
    date_obs : string
        The starting date, UTC.
    latitude : float
        The observer's latitude geolocation. Default is Dome C.
    longitude : float
        The observer's longitude geolocation. Default is Dome C.

    Example
    -------
    >>> gal2hor(0, 0, 0)
    (array(50.35837815921487), array(39.212362279976155))

    """
    incoords, time = _format_sphconv(l, b, date_obs, time)
    g2e = SphericalGalactic2EquatorialOperator(degrees=True)
    e2h = SphericalEquatorial2HorizontalOperator(
        'NE', time, latitude, longitude, degrees=True)
    outcoords = e2h(g2e(incoords))
    return outcoords[..., 0], outcoords[..., 1]


def hor2gal(azimuth, elevation, time, date_obs=QubicPointing.DEFAULT_DATE_OBS,
            latitude=DOMECLAT, longitude=DOMECLON):
    """
    hor2gal(az, el, time, [date_obs, [latitude, [longitude]]]) -> l, b
    Horizontal to galactic spherical conversion. Angles are in degrees.

    Parameters
    ----------
    time : array-like
        Elapsed time in seconds since date_obs.
    date_obs : string
        The starting date, UTC.
    latitude : float
        The observer's latitude geolocation. Default is Dome C.
    longitude : float
        The observer's longitude geolocation. Default is Dome C.

    Example
    -------
    >>> hor2gal(50.35837815921487, 39.212362279976155, 0)
    (array(4.452776554048925e-14), array(-7.63333123551244e-14))

    """
    incoords, time = _format_sphconv(azimuth, elevation, date_obs, time)
    h2e = SphericalHorizontal2EquatorialOperator(
        'NE', time, latitude, longitude, degrees=True)
    e2g = SphericalEquatorial2GalacticOperator(degrees=True)
    outcoords = e2g(h2e(incoords))
    return outcoords[..., 0], outcoords[..., 1]



from __future__ import division

import numpy as np
from itertools import izip
from pyoperators import (
    BlockColumnOperator, BlockDiagonalOperator, DiagonalOperator,
    MPIDistributionIdentityOperator, PackOperator, pcg, proxy_group,
    rule_manager)
from .acquisition import QubicAcquisition
from .utils import progress_bar

__all__ = ['map2tod', 'tod2map_all', 'tod2map_each']


def _get_projection_restricted(acq, P, mask):
    #XXX HACK
    if len(acq.block) == 1:
        P.restrict(mask)
        return P

    def callback(i):
        f = acq.instrument.get_projection_operator
        p = f(acq.sampling[acq.block[i]], acq.scene, verbose=False)
        p.restrict(mask)
        return p
    shapeouts = [(len(acq.instrument), s.stop-s.start) +
                 acq.scene.shape[1:] for s in acq.block]
    proxies = proxy_group(len(acq.block), callback, shapeouts=shapeouts)
    return BlockColumnOperator(proxies, axisout=1)

#    nbytes = acq.get_projection_nbytes()
#    if max_nbytes is None or nbytes <= max_nbytes:
#        if acq.comm.size > 1:
#            P = P.operands[0]
#        if isinstance(P, BlockColumnOperator):
#            for _ in P.operands:
#                _.restrict(mask)
#        else:
#            P.restrict(mask)
#        return P
#    n = int(np.ceil(nbytes / max_nbytes))
#    samplings = acq.sampling.split(n)
#    acqs = [QubicAcquisition(acq.instrument, _, acq.scene, acq.block)
#            for _ in samplings]
#
#    def callback(i):
#        p = acqs[i].get_projection_operator(verbose=False)
#        p.restrict(mask)
#        return p
#    return BlockColumnOperator(proxy_group(n, callback), axisout=1)


def map2tod(acq, map, convolution=False, max_nbytes=None):
    """
    tod = map2tod(acquisition, map)
    tod, convolved_map = map2tod(acquisition, map, convolution=True)

    Parameters
    ----------
    acq : QubicAcquisition
        The QUBIC acquisition.
    map : I, QU or IQU maps
        Temperature, QU or IQU maps of shapes npix, (npix, 2), (npix, 3)
        with npix = 12 * nside**2
    convolution : boolean, optional
        Set to True to convolve the input map by a gaussian and return it.
    max_nbytes : int
        Maximum number of bytes for the pointing matrix. If the actual size
        is greater than this number, the computation of the pointing matrix
        will be performed on the fly at each iteration.

    Returns
    -------
    tod : array
        The Time-Ordered-Data of shape (ndetectors, ntimes).
    convolved_map : array
        The convolved map, if the convolution keyword is set.

    """
    if convolution:
        convolution = acq.get_convolution_peak_operator()
        map = convolution(map)

    H = acq.get_operator()
    tod = H(map)

    if convolution:
        return tod, map

    return tod


def _tod2map(acq, tod, coverage_threshold, disp_pmatrix, max_nbytes, callback,
             disp_pcg, maxiter, tol, _factor):
    projection = acq.get_projection_operator(verbose=disp_pmatrix)
    distribution = acq.get_distribution_operator()
    P = projection * distribution

    shape = len(acq.instrument), len(acq.sampling)
    kind = acq.scene.kind
    if kind == 'I':
        ones = np.ones(shape)
    elif kind == 'IQU':
        ones = np.zeros(shape + (3,))
        ones[..., 0] = 1
    else:
        raise NotImplementedError()
    coverage = P.T(ones)
    if kind == 'IQU':
        coverage = coverage[..., 0]
    cov = coverage[coverage > 0]
    i = np.argsort(cov)
    cdf = np.cumsum(cov[i])
    j = np.argmax(cdf >= coverage_threshold * cdf[-1])
    threshold = cov[i[j]]
    mask = coverage >= threshold
    if acq.comm.rank == 0 and coverage_threshold > 0:
        print 'Total coverage:', cdf[-1]
        print 'Threshold coverage set to:', threshold
        print 'Fraction of rejected pixels:', 1 - np.sum(mask) / cov.size

    projection = _get_projection_restricted(acq, projection, mask)
    pack = PackOperator(mask, broadcast='rightward')
    hwp = acq.get_hwp_operator()
    polarizer = acq.get_polarizer_operator()
    response = acq.get_detector_response_operator()
    H = response * polarizer * (hwp * projection) * distribution
    if acq.scene.kind == 'QU':
        H = acq.get_subtract_grid_operator() * H

    if _factor is not None:
        coef = np.array([1, _factor, _factor])
        H = H * DiagonalOperator(1 / coef, broadcast='leftward')

    invNtt = acq.get_invntt_operator()
    preconditioner = DiagonalOperator(1/coverage[mask], broadcast='rightward')
    A = H.T * invNtt * H
    solution = pcg(A, H.T(invNtt(tod)), M=preconditioner, callback=callback,
                   disp=disp_pcg, maxiter=maxiter, tol=tol)
    if _factor is not None:
        solution['x'] *= coef
    output_map = pack.T(solution['x'])
    return output_map, coverage


def tod2map_all(acquisition, tod, coverage_threshold=0.01, max_nbytes=None,
                callback=None, disp=True, maxiter=300, tol=1e-4, _factor=None):
    """
    Compute map using all detectors.

    map, coverage = tod2map_all(acquisition, tod, [coverage_threshold,
                                max_nbytes, callback, disp, tol])

    Parameters
    ----------
    acquisition : QubicAcquisition
        The QUBIC acquisition.
    tod : array-like
        The Time-Ordered-Data of shape (ndetectors, ntimes).
    coverage_threshold : float, optional
        The low-coverage sky pixels whose cumulative coverage is below a
        fraction of the total coverage are rejected. This keyword speficies
        this fraction (between 0 and 1]).
    max_nbytes : int
        Maximum number of bytes for the pointing matrix. If the actual size
        is greater than this number, the computation of the pointing matrix
        will be performed on the fly at each iteration.
    callback : function, optional
        User-supplied function to call after each iteration.  It is called
        as callback(solver), where solver is an Algorithm instance.
    disp : boolean, optional
        Display of solver's iterations.
    maxiter : integer, optional
        Maximum number of iterations.  Iteration will stop after maxiter
        steps even if the specified tolerance has not been achieved.
    tol : float, optional
        Solver tolerance.

    Returns
    -------
    map : I, QU or IQU maps
        Temperature, QU or IQU maps of shapes npix, (npix, 2), (npix, 3)
        with npix = 12 * nside**2
    """
    return _tod2map(acquisition, tod, coverage_threshold, True, max_nbytes,
                    callback, disp, maxiter, tol, _factor)


def tod2map_each(acquisition, tod, coverage_threshold=0, max_nbytes=None,
                 callback=None, disp=True, maxiter=300, tol=1e-4, _factor=None):
    """
    Compute average map from each detector.

    map, coverage = tod2map_each(acquisition, tod, [coverage_threshold,
                                 max_nbytes, callback, disp, tol])

    Parameters
    ----------
    acquisition : QubicAcquisition
        The QUBIC acquisition.
    tod : array-like
        The Time-Ordered-Data of shape (ndetectors, ntimes).
    coverage_threshold : float, optional
        The coverage threshold used to reject map pixels from
        the reconstruction.
    max_nbytes : int
        Maximum number of bytes for the pointing matrix. If the actual size
        is greater than this number, the computation of the pointing matrix
        will be performed on the fly at each iteration.
    callback : function, optional
        User-supplied function to call after each iteration.  It is called
        as callback(solver), where solver is an Algorithm instance.
    disp : boolean, optional
        Display of solver's iterations.
    maxiter : integer, optional
        Maximum number of iterations.  Iteration will stop after maxiter
        steps even if the specified tolerance has not been achieved.
    tol : float, optional
        Solver tolerance.

    Returns
    -------
    map : I, QU or IQU maps
        Temperature, QU or IQU maps of shapes npix, (npix, 2), (npix, 3)
        with npix = 12 * nside**2

    """
    instrument = acquisition.instrument
    x = np.zeros(acquisition.scene.shape)
    n = np.zeros(acquisition.scene.shape[0])
    if disp:
        bar = progress_bar(len(instrument), 'TOD2MAP_EACH')
    for i, t in izip(instrument, tod):
        acq = type(acquisition)(i, acquisition.sampling, acquisition.scene)
        x_, n_ = _tod2map(acq, t[None, :], coverage_threshold, False,
                          max_nbytes, callback, False, maxiter, tol, _factor)
        x += x_
        n += n_
        if disp:
            bar.update()

    if acquisition.scene.kind == 'I':
        return np.nan_to_num(x / n), n
    return np.nan_to_num(x / n[:, None]), n



from __future__ import division

import healpy as hp
import matplotlib.pyplot as mp
import numpy as np
import os
import qubic
from pyoperators import DiagonalOperator, PackOperator, pcg
from pyoperators import rule_manager
from qubic import (
    QubicAcquisition, QubicInstrument, create_sweeping_pointings, equ2gal)

DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')

# acquisition parameters
nside = 256
racenter = 0.0      # deg
deccenter = -57.0   # deg
angspeed = 1        # deg/sec
delta_az = 15.      # deg
angspeed_psi = 0.1  # deg/sec
maxpsi = 45.        # deg
nsweeps_el = 300
duration = 24       # hours
ts = 0.1            # seconds

# get the sampling model
np.random.seed(0)
sampling = create_sweeping_pointings(
    [racenter, deccenter], duration, ts, angspeed, delta_az, nsweeps_el,
    angspeed_psi, maxpsi)
sampling.angle_hwp = np.random.random_integers(0, 7, len(sampling)) * 11.25

# get the instrument model with only one detector
idetector = 0
instrument = QubicInstrument()[idetector]

# get the acquisition model from the instrument and sampling models
acq = QubicAcquisition(instrument, sampling, nside=nside)

# get it as operators
convolution = acq.get_convolution_peak_operator()
projection = acq.get_projection_operator()
hwp = acq.get_hwp_operator()
polarizer = acq.get_polarizer_operator()

# restrict the projection to the observed sky pixels
coverage = projection.pT1()
mask = coverage > 0
projection.restrict(mask)
pack = PackOperator(mask, broadcast='rightward')

with rule_manager(inplace=True):
    H = polarizer * (hwp * projection)

# get noiseless timeline
x0 = qubic.io.read_map(DATAPATH)
x0_convolved = convolution(x0)
y = H(pack(x0_convolved))

# inversion through Preconditioned Conjugate Gradient
preconditioner = DiagonalOperator(1/coverage[mask], broadcast='rightward')
solution = pcg(H.T * H, H.T(y), M=preconditioner, disp=True, tol=1e-3)
output_map = pack.T(solution['x'])


# some display
def display(input, msg):
    for i, (kind, lim) in enumerate(zip('IQU', [200, 10, 10])):
        map = input[..., i].copy()
        map[~mask] = np.nan
        hp.gnomview(map, rot=center, reso=5, xsize=400, min=-lim, max=lim,
                    title=msg + ' ' + kind, sub=(1, 3, i+1))

center = equ2gal(racenter, deccenter)

mp.figure(1)
display(x0_convolved, 'Original map')
mp.figure(2)
display(output_map, 'Reconstructed map')
mp.figure(3)
display(output_map-x0_convolved, 'Difference map')
mp.show()



from . import io
from .acquisition import *
from .beams import *
from .calibration import *
from .cl import *
from .instrument import *
from .mapmaking import *
from .pointings import *
from .scene import *

__version__ = '0.4.1'



from __future__ import absolute_import, division, print_function
import warnings
from warnings import warn


class QubicWarning(UserWarning):
    pass


class QubicDeprecationWarning(DeprecationWarning):
    pass

warnings.simplefilter('always', category=QubicWarning)
warnings.simplefilter('module', category=QubicDeprecationWarning)
del warnings



"""
Noiseless simulation and map-making using the gaussian approximation
for the synthetic beam.

"""
from __future__ import division

import healpy as hp
import matplotlib.pyplot as mp
import numpy as np
import os
import qubic
from pyoperators import pcg, DiagonalOperator, UnpackOperator
from qubic import QubicAcquisition, create_random_pointings, gal2equ

# read the input map
DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
input_map = qubic.io.read_map(DATAPATH, field='I_STOKES')

# let's take the galactic north pole as the center of the observation field
center_gal = 0, 90
center = gal2equ(center_gal[0], center_gal[1])

# sampling model
np.random.seed(0)
sampling = create_random_pointings(center, 1000, 10)

# acquisition model
acq = QubicAcquisition(150, sampling, kind='I')
hit = acq.get_hitmap()

C = acq.get_convolution_peak_operator()
P = acq.get_projection_operator()
H = P * C

# Produce the Time-Ordered data
tod = H(input_map)
#acq.save_simulation('mysimul', input_map, tod, 'my info')

# map-making
coverage = P.pT1()
mask = coverage > 10
P.restrict(mask)
unpack = UnpackOperator(mask)
solution = pcg(P.T * P, P.T(tod),
               M=DiagonalOperator(1/coverage[mask]), disp=True)
output_map = unpack(solution['x'])


# some display
def display(x, title):
    x = x.copy()
    x[~mask] = np.nan
    hp.gnomview(x, rot=center_gal, reso=5, xsize=600, min=-200, max=200,
                title=title)

display(input_map, 'Original map')
display(C(input_map), 'Convolved original map')
display(output_map, 'Reconstructed map (gaussian approximation)')
mp.show()



from __future__ import division

import numpy as np
try:
    import matplotlib.pyplot as mp
except:
    pass
import os

__all__ = ['read_spectra', 'plot_spectra']


def read_spectra():
    """
    Read a file containing Cl spectra for r=0.1.

    Returns
        C_TT, C_EE, C_BB, C_TE : 4-tuple of ndarrays
        The Cl spectra for the T-T, E-E, T-E and T-E correlations.

    """
    filename = os.path.join(os.path.dirname(__file__),
                            'data', 'cl_r=0.1bis2.txt')
    # in this file, the spectra start from l=2
    ell, b, c, d, e, f, g, h, i, j, k = np.loadtxt(filename, unpack=True)
    ctt = np.r_[0, 0, b * 2e12 * np.pi / (ell * (ell + 1))]
    cee = np.r_[0, 0, c * 2e12 * np.pi / (ell * (ell + 1))]
    cte = np.r_[0, 0, e * 2e12 * np.pi / (ell * (ell + 1))]
    cbb = np.r_[0, 0, h * 2e12 * np.pi / (ell * (ell + 1))]
    return ctt, cee, cbb, cte


def plot_spectra(spectra, label_bb='$C_\ell^{BB}$'):
    """
    Plot C_TT, C_EE, C_BB and C_TE spectra.

    Parameters
    ----------
    spectra : 4-tuple
        The C_TT, C_EE, C_BB and C_TE spectra (starting from l=0).

    """
    ctt, cee, cbb, cte = spectra
    ell = np.arange(ctt.size)
    norm = (ell * (ell + 1)) / (2 * np.pi)
    mp.semilogy(ell, np.sqrt(norm * ctt), 'k', label='$C_\ell^{TT}$')
    mp.plot(ell, np.sqrt(norm * cte), 'g', label='$C_\ell^{TE}$')
    mp.plot(ell, np.sqrt(-norm * cte), 'g--',)
    mp.plot(ell, np.sqrt(norm * cee), 'b', label='$C_\ell^{EE}$')
    mp.plot(ell, np.sqrt(norm * cbb), color='brown', label=label_bb)
    mp.xlim(0, 600)
    mp.ylim(0.01, 100)
    mp.xlabel('$\ell$')
    mp.ylabel('$\sqrt{\ell(\ell+1)|C_\ell|/2\pi}$    $[\mu K]$')
    mp.legend(loc='best', frameon=False)



"""
Noiseless simulation and map-making using the gaussian approximation
for the synthetic beam.

"""
from __future__ import division

import healpy as hp
import matplotlib.pyplot as mp
import numpy as np
import os
import qubic
from qubic import (
    create_random_pointings, gal2equ, map2tod, QubicAcquisition,
    SimpleAcquisition, tod2map_all)

# read the input map
DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
input_map = qubic.io.read_map(DATAPATH, field='I_STOKES')

# let's take the galactic north pole as the center of the observation field
center_gal = 0, 90
center = gal2equ(center_gal[0], center_gal[1])

# sampling model
np.random.seed(0)
sampling = create_random_pointings(center, 1000, 10)

# acquisition model
acq = QubicAcquisition(150, sampling, kind='I')
acq_simple = SimpleAcquisition(150, sampling, kind='I')
hit = acq.get_hitmap()
C = acq.get_convolution_peak_operator()

tod = map2tod(acq, input_map)
output_map, coverage = tod2map_all(acq, tod)
output_map[coverage == 0] = np.nan
coverage[coverage == 0] = np.nan

tod_simple = map2tod(acq_simple, input_map)
output_map_simple, coverage_simple = tod2map_all(acq_simple, tod_simple)
output_map_simple[coverage_simple == 0] = np.nan
coverage_simple[coverage_simple == 0] = np.nan


# some display
def display(x, title, min=-200, max=200):
    hp.gnomview(x, rot=center_gal, reso=5, xsize=600, min=min, max=max,
                title=title)

display(input_map, 'Original map')
display(C(input_map), 'Convolved original map')
display(output_map, 'Reconstructed map (gaussian approximation)')
display(output_map_simple, 'Reconstructed map (simple model)')
maxval = coverage_simple.max()
display(coverage, 'Coverage', min=0, max=maxval)
display(coverage_simple, 'Coverage (simple model)', min=0, max=maxval)
mp.show()



from __future__ import division

from pyoperators.utils import settingerr
from pyoperators.utils.testing import assert_equal, assert_same, assert_is_type
from pysimulators import FitsArray
from qubic import create_random_pointings, QubicAcquisition, QubicInstrument
from qubic.beams import GaussianBeam, UniformHalfSpaceBeam

q = QubicInstrument(detector_ngrids=1)
s = create_random_pointings([0, 90], 5, 10.)


def test_detector_indexing():
    expected = FitsArray('test/data/detector_indexing.fits')
    assert_same(q.detector.index, expected)


def test_beams():
    assert_is_type(q.primary_beam, GaussianBeam)
    assert_is_type(q.secondary_beam, GaussianBeam)
    a = QubicAcquisition(150, s, primary_beam=UniformHalfSpaceBeam(),
                         secondary_beam=UniformHalfSpaceBeam())
    assert_is_type(a.instrument.primary_beam, UniformHalfSpaceBeam)
    assert_is_type(a.instrument.secondary_beam, UniformHalfSpaceBeam)


def test_primary_beam():
    def primary_beam(theta, phi):
        import numpy as np
        with settingerr(invalid='ignore'):
            return theta <= np.radians(9)
    a = QubicAcquisition(150, s, primary_beam=primary_beam)
    p = a.get_projection_operator()
    assert_equal(p.matrix.ncolmax, 5)



from __future__ import division

import healpy as hp
import matplotlib.pyplot as mp
import numpy as np
import os
import qubic
from pyoperators import pcg, DiagonalOperator, UnpackOperator
from pysimulators import SphericalEquatorial2GalacticOperator
from pysimulators.noises import _gaussian_psd_1f
from qubic import (
    QubicAcquisition, QubicInstrument, create_sweeping_pointings, equ2gal)

# read the input map
DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
input_map = qubic.io.read_map(DATAPATH, field='I_STOKES')


# instrument model
sigma = 10
fknee = 1
fslope = 1
ncorr = 10

# observation model
np.random.seed(0)
racenter = 0.0
deccenter = -57.0
angspeed = 1  # deg/sec
delta_az = 15.
angspeed_psi = 0.1
maxpsi = 45.
nsweeps_el = 300
duration = 24   # hours
ts = 20         # seconds
sampling = create_sweeping_pointings(
    [racenter, deccenter], duration, ts, angspeed, delta_az, nsweeps_el,
    angspeed_psi, maxpsi)

# acquisition model
acq = QubicAcquisition(150, sampling, kind='I', synthbeam_fraction=0.99,
                       detector_sigma=sigma, detector_fknee=fknee,
                       detector_fslope=fslope, detector_ncorr=ncorr)
C = acq.get_convolution_peak_operator()
P = acq.get_projection_operator()
H = P * C

# produce the Time-Ordered data
tod = H(input_map)

# noise
psd = _gaussian_psd_1f(len(acq.sampling), sigma=sigma, fknee=fknee,
                       fslope=fslope, sampling_frequency=1/ts)
invntt = acq.get_invntt_operator()
noise = acq.get_noise()

# map-making
coverage = P.pT1()
mask = coverage > 10
P.restrict(mask)
unpack = UnpackOperator(mask)

# map without covariance matrix
solution1 = pcg(P.T * P, P.T(tod + noise),
                M=DiagonalOperator(1/coverage[mask]), disp=True)
output_map1 = unpack(solution1['x'])

# map with covariance matrix
solution2 = pcg(P.T * invntt * P, (P.T * invntt)(tod + noise),
                M=DiagonalOperator(1/coverage[mask]), disp=True)
output_map2 = unpack(solution2['x'])

center = equ2gal(racenter, deccenter)


def display(x, title):
    x = x.copy()
    x[~mask] = np.nan
    hp.gnomview(x, rot=center, reso=5, xsize=600, min=-200, max=200,
                title=title)

display(C(input_map), 'Original convolved map')
#hp.projplot(np.radians(pointings[..., 0]), np.radians(pointings[..., 1]))
display(output_map1, 'Reconstructed map no invntt')
display(output_map2, 'Reconstructed map with invntt')
display(coverage, 'Coverage')

mp.show()



from __future__ import absolute_import, division, print_function
from pysimulators.interfaces.healpy import SceneHealpix

__all__ = ['QubicScene']


class QubicScene(SceneHealpix):
    def __init__(self, band, nside=256, kind='IQU'):
        """
        Parameters
        ----------
        band : int
            The operating frequency, in GHz.
        nside : int
            The Healpix scene's nside.
        kind : 'I', 'QU', 'IQU'
            The sky kind: 'I' for intensity-only, 'QU' for Q and U maps,
            and 'IQU' for intensity plus QU maps.

        """
        self.nu = band * 1e9
        self.monochromatic = True
        SceneHealpix.__init__(self, nside, kind=kind)



from __future__ import division

import numpy as np
import os
import qubic
from numpy.testing import assert_allclose, assert_equal
from pyoperators import pcg, DiagonalOperator, UnpackOperator
from pyoperators.utils.testing import assert_same
from qubic import QubicAcquisition, QubicInstrument, create_sweeping_pointings

# read the input map
DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
I, Q, U = qubic.io.read_map(DATAPATH,
                            field=('I_STOKES', 'Q_STOKES', 'U_STOKES')).T

# observation model
np.random.seed(0)
racenter = 0.0
deccenter = -57.0
angspeed = 1  # deg/sec
delta_az = 15.
angspeed_psi = 0.1
maxpsi = 45.
nsweeps_el = 300
duration = 1   # hours
ts = 20        # seconds
pointings = create_sweeping_pointings(
    [racenter, deccenter], duration, ts, angspeed, delta_az, nsweeps_el,
    angspeed_psi, maxpsi)

tods = {}
pTxs = {}
pT1s = {}
pTx_pT1s = {}
cbiks = {}
outputs = {}
kinds = ['I', 'QU', 'IQU']
input_maps = [I,
              np.array([Q, U]).T,
              np.array([I, Q, U]).T]
for kind, input_map in zip(kinds, input_maps):
    acq = QubicAcquisition(150, pointings, kind=kind, ngrids=2)
    P = acq.get_projection_operator()
    W = acq.get_hwp_operator()
    H = W * P
    coverage = P.pT1()
    tod = H(input_map)
    tods[kind] = tod
    pTxs[kind] = H.T(tod)[coverage > 0]
    if kind != 'QU':
        pTx_pT1 = P.pTx_pT1(tod)
        pTx_pT1s[kind] = pTx_pT1[0][coverage > 0], pTx_pT1[1][coverage > 0]
    cbik = P.canonical_basis_in_kernel()
    mask = coverage > 10
    P.restrict(mask)
    unpack = UnpackOperator(mask, broadcast='rightward')
    x0 = unpack.T(input_map)

    M = DiagonalOperator(1 / coverage[mask], broadcast='rightward')

    H = W * P
    solution = pcg(H.T * H, H.T(tod), M=M, disp=True, tol=1e-5)
    pT1s[kind] = coverage
    cbiks[kind] = cbik
    outputs[kind] = solution['x']

assert_same(tods['I'], tods['IQU'][..., 0])
assert_same(tods['QU'], tods['IQU'][..., 1:])

assert_same(pTxs['I'], pTxs['IQU'][..., 0].astype(np.float32))
assert_same(pTxs['QU'], pTxs['IQU'][..., 1:].astype(np.float32))

assert_same(pTx_pT1s['I'][0], pTx_pT1s['IQU'][0].astype(np.float32))
assert_same(pTx_pT1s['I'][1], pTx_pT1s['IQU'][1].astype(np.float32))

assert_allclose(outputs['I'], outputs['IQU'][..., 0], atol=2e-2)
assert_allclose(outputs['QU'], outputs['IQU'][..., 1:], atol=2e-2)

for k in ('QU', 'IQU'):
    assert_equal(cbiks[k], cbiks['I'])
    assert_same(pT1s[k], pT1s['I'].astype(np.float32), rtol=15)



from __future__ import division

import numpy as np
from pyoperators.utils.testing import assert_same
from qubic.beams import GaussianBeam, UniformHalfSpaceBeam

angles = np.radians([0, 10, 20, 30, 40, 50, 60, 70, 80, 90])


def test_gaussian():
    primary = GaussianBeam(14)
    secondary = GaussianBeam(14, backward=True)
    assert_same(primary(np.pi - angles), secondary(angles))


def test_uniform():
    beam = UniformHalfSpaceBeam()
    assert_same(beam(10), 1)



from __future__ import division

import numpy as np
from progressbar import ProgressBar, Bar, ETA, Percentage


class _ProgressBar(ProgressBar):
    def update(self, n=None):
        if n is not None:
            ProgressBar.update(self, n)
            return
        ProgressBar.update(self, self.currval + 1)
        if self.currval >= self.maxval:
            self.finish()


def progress_bar(n, info=''):
    """
    Return a default progress bar.

    Example
    -------
    >>> import time
    >>> n = 10
    >>> bar = progress_bar(n, 'LOOP')
    >>> for i in range(n):
    ...     time.sleep(1)
    ...     bar.update()

    """
    return _ProgressBar(widgets=[info, Percentage(), Bar('=', '[', ']'),
                                 ETA()], maxval=n).start()


def _compress_mask(mask):
    mask = mask.ravel()
    if len(mask) == 0:
        return ''
    output = ''
    old = mask[0]
    n = 1
    for new in mask[1:]:
        if new is not old:
            if n > 2:
                output += str(n)
            elif n == 2:
                output += '+' if old else '-'
            output += '+' if old else '-'
            n = 1
            old = new
        else:
            n += 1
    if n > 2:
        output += str(n)
    elif n == 2:
        output += '+' if old else '-'
    output += '+' if old else '-'
    return output


def _uncompress_mask(mask):
    i = 0
    l = []
    nmask = len(mask)
    while i < nmask:
        val = mask[i]
        if val == '+':
            l.append(True)
            i += 1
        elif val == '-':
            l.append(False)
            i += 1
        else:
            j = i + 1
            val = mask[j]
            while val not in ('+', '-'):
                j += 1
                val = mask[j]
            l.extend(int(mask[i:j]) * (True if val == '+' else False,))
            i = j + 1
    return np.array(l, bool)



"""
Noiseless simulation and map-making using the gaussian approximation
for the synthetic beam.

"""
from __future__ import division
import numpy as np
import os
import qubic
from qubic import (
    QubicAcquisition, create_random_pointings, gal2equ, map2tod, tod2map_all)

# read the input map
DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data',
                        'syn256_pol.fits')
input_map = qubic.io.read_map(DATAPATH, field='I_STOKES')

# let's take the galactic north pole as the center of the observation field
center_gal = 0, 90
center = gal2equ(center_gal[0], center_gal[1])

# sampling model
np.random.seed(0)
sampling = create_random_pointings(center, 1000, 10)

# acquisition model
acq = QubicAcquisition(150, sampling, kind='I')
hit = acq.get_hitmap()

# Produce the Time-Ordered data
tod = map2tod(acq, input_map)
output_map, coverage = tod2map_all(acq, tod)

print acq.comm.rank, output_map[coverage > 0][:5]
print acq.comm.rank, hit[hit > 0][:10]



from __future__ import division

import numpy as np
from numpy import pi

__all__ = []


class Beam(object):
    def __init__(self, solid_angle):
        """
        Parameter
        ---------
        solid_angle : float
            The beam solid angle [sr].
        """
        self.solid_angle = float(solid_angle)

    def __call__(self, theta_rad, phi_rad):
        raise NotImplementedError()


class GaussianBeam(Beam):
    """
    Axisymmetric gaussian beam.

    """
    def __init__(self, fwhm_deg, backward=False):
        self.sigma_deg = fwhm_deg / np.sqrt(8 * np.log(2))
        self.sigma_rad = np.radians(self.sigma_deg)
        self.fwhm_deg = fwhm_deg
        self.backward = bool(backward)
        Beam.__init__(self, 2 * pi * self.sigma_rad**2)

    def __call__(self, theta_rad, phi_rad=None):
        if self.backward:
            theta_rad = pi - theta_rad
        return np.exp(-theta_rad**2 / (2 * self.sigma_rad**2))


class UniformHalfSpaceBeam(Beam):
    """
    Uniform beam in half-space.

    """
    def __init__(self):
        Beam.__init__(self, 2 * pi)

    def __call__(self, theta_rad, phi_rad=None):
        return 1



from __future__ import division

import healpy as hp
import numpy as np
import tempfile
from numpy.testing import assert_equal, assert_raises
from pyoperators.utils.testing import assert_is_none, assert_same
from qubic.io import read_map, write_map

dtypes = (np.uint8, np.int16, np.int32, np.int64, np.float32, np.float64)
nside = 16
npix = 12 * nside**2


def test_write_read():
    complete = np.arange(npix, dtype=float)
    complete[npix // 10:] = np.nan
    mask = np.isfinite(complete)
    partial = complete[mask]

    def func(map, msk, partial, compress):
        with tempfile.TemporaryFile('a+b') as tmpfile:
            write_map(tmpfile, map, msk, compress=compress)
            tmpfile.seek(0)
            actual = read_map(tmpfile, partial=partial)
        if partial:
            out, mask_ = actual
            if msk is None:
                assert_is_none(mask_)
                actual = out
            else:
                actual = np.full(mask_.shape, np.nan)
                actual[mask_] = out
        assert_equal(actual.dtype.byteorder, '=')
        assert_same(actual, complete)
    for map, msk in (complete, None), (partial, mask):
        for partial in False, True:
            for compress in False, True:
                yield func, map, msk, partial, compress


def test_write_read_field():
    complete = np.arange(3*npix, dtype=float).reshape(npix, 3)
    mask = np.zeros(npix, bool)
    mask[:npix // 10] = True
    complete[~mask] = np.nan
    partial = complete[mask]
    fields = 0, (2,), (0, 1), (1, 2), (0, 1, 2), (0, 1, 0)

    def func(dtype, field, map, msk):
        with tempfile.TemporaryFile('a+b') as tmpfile:
            write_map(tmpfile, map.astype(dtype), msk, dtype=None)
            tmpfile.seek(0)
            actual = read_map(tmpfile, field=field, dtype=None)
        assert actual.dtype == dtype
        if np.isscalar(field):
            assert_equal(actual, complete[..., field].astype(dtype))
        elif len(field) == 1:
            assert_equal(actual, complete[..., field[0]].astype(dtype))
        else:
            for i, f in enumerate(field):
                assert_equal(actual[..., i], complete[..., f].astype(dtype))
    for map, msk in (complete, None), (partial, mask):
        for dtype in dtypes:
            for field in fields:
                yield func, dtype, field, map, msk


def test_read_header():
    with tempfile.TemporaryFile('a+b') as tmpfile:
        write_map(tmpfile, np.arange(12), coord='E')
        tmpfile.seek(0)
        actual, header = read_map(tmpfile, h=True)
    assert_equal(header['coordsys'], 'E')
    assert_equal(header['nmaps'], 1)
    assert not header['hasmask']


def test_read_healpy():
    def func(h, partial):
        with tempfile.NamedTemporaryFile('a+b') as tmpfile:
            hp.write_map(tmpfile.name, (np.arange(12), np.arange(12) + 1),
                         coord='E')
            actual = read_map(tmpfile.name, h=h, partial=partial)
            expected = hp.read_map(tmpfile.name, h=h, field=(0, 1))
        if partial:
            assert_is_none(actual[1])
            if h:
                actual = actual[0], actual[2]
            else:
                actual = actual[0]
        if h:
            assert actual[1].items() == expected[-1]
            assert_equal(actual[0], np.column_stack(expected[:-1]))
        else:
            assert_equal(actual, np.column_stack(expected))
    for h in False, True:
        for partial in False, True:
            yield func, h, partial


def test_read_nest():
    def func(nwrite, nread):
        with tempfile.TemporaryFile('a+b') as tmpfile:
            write_map(tmpfile, np.arange(12*16), nest=nwrite)
            tmpfile.seek(0)
            actual = read_map(tmpfile, nest=nread)
        with tempfile.NamedTemporaryFile('a+b') as tmpfile:
            hp.write_map(tmpfile.name, np.arange(12*16), nest=nwrite)
            expected = hp.read_map(tmpfile.name, nest=nread)
        assert_equal(expected, actual)
    for nwrite in False, True:
        for nread in False, True:
            yield func, nwrite, nread


def test_write_errors():
    assert_raises(ValueError, write_map, 'test.fits', np.array(3))
    assert_raises(ValueError, write_map, 'test.fits', np.zeros((1, 1, 1)))
    assert_raises(ValueError, write_map, 'test.fits', np.zeros(12), 3)
    assert_raises(ValueError, write_map, 'test.fits', np.zeros(13))



from __future__ import division

import numpy as np
from astropy.time import Time
from pyoperators.utils.testing import assert_same
from pysimulators import (
    SphericalEquatorial2GalacticOperator,
    SphericalEquatorial2HorizontalOperator)
from qubic.pointings import (
    equ2gal, gal2equ, equ2hor, hor2equ, gal2hor, hor2gal, DOMECLAT, DOMECLON,
    QubicPointing)


def test_sphconv():

    time = Time(QubicPointing.DEFAULT_DATE_OBS, scale='utc')

    def _pack(x):
        return np.array([x[0], x[1]]).T

    sphs = equ2gal, gal2equ, equ2hor, hor2equ, gal2hor, hor2gal
    extraargs = (), (), (0,), (0,), (0,), (0,)
    e2g = SphericalEquatorial2GalacticOperator(degrees=True)
    e2h = SphericalEquatorial2HorizontalOperator(
        'NE', time, DOMECLAT, DOMECLON, degrees=True)
    refs = e2g, e2g.I, e2h, e2h.I, e2h(e2g.I), e2g(e2h.I)

    incoords = np.array([[10, 20], [30, 40], [50, 60]])

    def func(sph, extraarg, ref):
        args = (incoords[..., 0], incoords[..., 1]) + extraarg
        outcoords = _pack(sph(*args))
        assert_same(outcoords, ref(incoords), rtol=100)

    for sph, extraarg, ref in zip(sphs, extraargs, refs):
        yield func, sph, extraarg, ref



"""
The version number is obtained from git tags, branch and commit identifier.
It has been designed for the following workflow:

- git checkout master
- modify, commit, commit
- set version 0.1 in setup.py -> 0.1.dev03
- modify, commit              -> 0.1.dev04
- git checkout -b v0.1        -> 0.1.dev04
- modify, commit              -> 0.1.pre01
- modify, commit              -> 0.1.pre02
- git tag 0.1                 -> 0.1
- modify... and commit        -> 0.1.post01
- modify... and commit        -> 0.1.post02
- git tag 0.1.1               -> 0.1.1
- modify... and commit        -> 0.1.1.post01
- git checkout master         -> 0.1.dev04
- set version=0.2 in setup.py -> 0.2.dev01
- modify, commit              -> 0.1.dev02

When working on the master branch, the dev number is the number of commits
since the last branch of name "v[0-9.]+"

"""
import os
import re
import sys
from numpy.distutils.command.build import build
from numpy.distutils.command.build_ext import build_ext
from numpy.distutils.command.sdist import sdist
from numpy.distutils.core import Command
from subprocess import call, Popen, PIPE
from warnings import filterwarnings

try:
    root = os.path.dirname(os.path.abspath(__file__))
except NameError:
    root = os.path.dirname(os.path.abspath(sys.argv[0]))
ABBREV = 5
BRANCH_REGEX = '^refs/(heads|remotes/origin)/v[0-9.]+$'


def get_version(name, default):
    version = get_version_git(default)
    if version != '':
        return version
    return get_version_init_file(name) or default


def get_version_git(default):
    def run(cmd, cwd=root):
        git = "git"
        if sys.platform == "win32":
            git = "git.cmd"
        process = Popen([git] + cmd, cwd=cwd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        if stderr != '':
            raise RuntimeError(stderr)
        if process.returncode != 0:
            raise RuntimeError('Error code: {0}.'.format(process.returncode))
        return stdout.strip()

    def get_branches():
        return run(['for-each-ref', '--sort=-committerdate', '--format=%(ref'
                    'name)', 'refs/heads', 'refs/remotes/origin']).split('\n')

    def get_branch_name():
        return run(['rev-parse', '--abbrev-ref', 'HEAD'])

    def get_description():
        try:
            description = run([
                'describe', '--tags', '--abbrev={0}'.format(ABBREV)])
        except RuntimeError:
            description = run([
                'describe', '--tags', '--abbrev={0}'.format(ABBREV),
                '--always']).split('-')
            return '', '', description[0], '-' + description[1]
        regex = r"""^
        (?P<tag>.*?)
        (?:-
            (?P<rev>\d+)-g
            (?P<commit>[0-9a-f]{5,40})
        )?
        $"""
        m = re.match(regex, description, re.VERBOSE)
        tag, rev, commit = (m.group(_) for _ in 'tag,rev,commit'.split(','))
        rev = int(rev)
        return tag, rev, commit

    def get_rev_since_branch(branch):
        common = run(['merge-base', 'HEAD', branch])
        return int(run(['rev-list', '--count', 'HEAD', '^' + common]))

    def get_dirty():
        return '-dirty' if run(['diff-index', 'HEAD']) else ''

    def get_master_rev(default):
        branches = get_branches()
        for branch in branches:
            # filter branches according to BRANCH_REGEX
            if not re.match(BRANCH_REGEX, branch):
                continue
            rev = get_rev_since_branch(branch)
            if rev > 0:
                return rev
        return int(run(['rev-list', '--count', 'HEAD']))

    try:
        run(['rev-parse', '--is-inside-work-tree'])
    except (OSError, RuntimeError):
        return ''

    dirty = get_dirty()

    # check if HEAD is tagged
    try:
        return run(['describe', '--tags', '--candidates=0']) + dirty
    except RuntimeError:
        pass

    # if the current branch is master, look up the last release branch
    # to get the dev number
    branch = get_branch_name()
    if get_branch_name() == 'master':
        rev = get_master_rev(default)
        commit = run(['rev-parse', '--short={}'.format(ABBREV), 'HEAD'])
        if default != '':
            return '{}.dev{:02}-g{}{}'.format(default, rev, commit, dirty)
        return str(rev) + dirty

    isrelease = re.match('^v[0-9.]+$', branch) is not None
    rev_master = get_rev_since_branch('master')
    tag, rev_tag, commit = get_description()
    if isrelease:
        version = tag
    else:
        version = branch
    if rev_tag > 0:
        if rev_master < rev_tag:
            version += '.pre{:02}'.format(rev_master)
        else:
            version += '.post{:02}'.format(rev_tag)
        version += '-g' + commit
    return version + dirty


def get_version_init_file(name):
    try:
        f = open(os.path.join(name, '__init__.py')).read()
    except IOError:
        return ''
    m = re.search(r"__version__ = '(.*)'", f)
    if m is None:
        return ''
    return m.groups()[0]


def write_version(name, version):
    try:
        init = open(os.path.join(root, name, '__init__.py.in')).readlines()
    except IOError:
        return
    init += ['\n', '__version__ = ' + repr(version) + '\n']
    open(os.path.join(root, name, '__init__.py'), 'w').writelines(init)


class BuildCommand(build):
    def run(self):
        write_version(self.distribution.get_name(),
                      self.distribution.get_version())
        build.run(self)


class SDistCommand(sdist):
    def make_release_tree(self, base_dir, files):
        write_version(self.distribution.get_name(),
                      self.distribution.get_version())
        initfile = os.path.join(self.distribution.get_name(), '__init__.py')
        if initfile not in files:
            files.append(initfile)
        sdist.make_release_tree(self, base_dir, files)


class CoverageCommand(Command):
    description = "run the package coverage"
    user_options = [('file=', 'f', 'restrict coverage to a specific file'),
                    ('erase', None,
                     'erase previously collected coverage before run'),
                    ('html-dir=', None,
                     'Produce HTML coverage information in dir')]

    def run(self):
        cmd = ['nosetests', '--with-coverage', '--cover-html',
               '--cover-package=' + self.distribution.get_name(),
               '--cover-html-dir=' + self.html_dir]
        if self.erase:
            cmd.append('--cover-erase')
        call(cmd + [self.file])

    def initialize_options(self):
        self.file = 'test'
        self.erase = 0
        self.html_dir = 'htmlcov'

    def finalize_options(self):
        pass


class TestCommand(Command):
    description = "run the test suite"
    user_options = [('file=', 'f', 'restrict test to a specific file')]

    def run(self):
        call(['nosetests', self.file])

    def initialize_options(self):
        self.file = 'test'

    def finalize_options(self):
        pass


def get_cmdclass():
    return {'build': BuildCommand,
            'build_ext': build_ext,
            'coverage': CoverageCommand,
            'sdist': SDistCommand,
            'test': TestCommand}

filterwarnings('ignore', "Unknown distribution option: 'install_requires'")



from __future__ import division

import matplotlib.pyplot as mp
import numexpr as ne
import numpy as np
from scipy.constants import c, pi
from pyoperators import MaskOperator
from pysimulators import create_fitsheader, SceneGrid
from qubic import QubicInstrument

qubic = QubicInstrument(detector_ngrids=1)

NU = 150e9                   # [Hz]
LAMBDA = c / NU              # [m]
FOCAL_LENGTH = 0.3           # [m]

SOURCE_POWER = 1             # [W]
SOURCE_THETA = np.radians(0) # [rad]
SOURCE_PHI = np.radians(45)  # [rad]

HORN_OPEN = np.zeros(qubic.horn.shape, dtype=bool)  # all closed
HORN_OPEN[10,10] = True
HORN_OPEN[11,11] = True
HORN_OPEN[0,1] = True
HORN_OPEN[17,18] = True
HORN_OPEN = ~qubic.horn.all.removed  # all open

DETECTOR_OFFSET = [0, 0]  # [m]

NPOINT_DETECTOR_PLANE = 512**2 # number of detector plane sampling points
DETECTOR_PLANE_LIMITS = (np.nanmin(qubic.detector.vertex[..., 0]),
                         np.nanmax(qubic.detector.vertex[..., 0]))  # [m]

# to check energy conservation (unrealistic detector plane):
#DETECTOR_PLANE_LIMITS = (-4, 4) # [m]

def ang2vec(theta_rad, phi_rad):
    sintheta = np.sin(theta_rad)
    return np.array([sintheta * np.cos(phi_rad),
                     sintheta * np.sin(phi_rad),
                     np.cos(theta_rad)])


########
# HORNS
########
nhorn_open = np.sum(HORN_OPEN)
horn_vec = np.column_stack([qubic.horn.all.center[HORN_OPEN],
                            np.zeros(nhorn_open)])


#################
# DETECTOR PLANE
#################
def norm(x):
    return x / np.sqrt(np.sum(x**2, axis=-1))[..., None]
ndet_x = int(np.sqrt(NPOINT_DETECTOR_PLANE))
a = np.r_[DETECTOR_PLANE_LIMITS[0]:DETECTOR_PLANE_LIMITS[1]:ndet_x*1j]
det_x, det_y = np.meshgrid(a, a)
det_spacing = (DETECTOR_PLANE_LIMITS[1] - DETECTOR_PLANE_LIMITS[0]) / ndet_x
det_vec = np.dstack([det_x, det_y, np.zeros_like(det_x) - FOCAL_LENGTH])
det_uvec = norm(det_vec)
det_theta = np.arccos(det_uvec[..., 2])

# solid angle of a detector plane pixel (gnomonic projection)
central_pixel_sr = np.arctan(det_spacing / FOCAL_LENGTH)**2
detector_plane_pixel_sr = -central_pixel_sr * np.cos(det_theta)**3


############
# DETECTORS
############
qubic.detector.vertex += DETECTOR_OFFSET
header = create_fitsheader((ndet_x, ndet_x), cdelt=det_spacing, crval=(0, 0),
                           ctype=['X---CAR', 'Y---CAR'], cunit=['m', 'm'])
detector_plane = SceneGrid.fromfits(header)
integ = MaskOperator(qubic.detector.all.removed) * \
        detector_plane.get_integration_operator(
            detector_plane.topixel(qubic.detector.all.vertex))


########
# MODEL
########
def get_model_A():
    """ Phase and transmission from the switches to the focal plane. """
    transmission = np.sqrt(qubic.secondary_beam(det_theta) *
                           detector_plane_pixel_sr /
                           qubic.secondary_beam.solid_angle)[..., None]
    const = 2j * pi / LAMBDA
    product = np.dot(det_uvec, horn_vec.T)
    return ne.evaluate('transmission * exp(const * product)')


def get_model_B():
    """ Phase and transmission from the source to the switches. """
    source_uvec = ang2vec(SOURCE_THETA, SOURCE_PHI)
    source_E = np.sqrt(SOURCE_POWER) * \
               np.sqrt(qubic.primary_beam(SOURCE_THETA, SOURCE_PHI))
    const = 2j * pi / LAMBDA
    product = np.dot(horn_vec, source_uvec)
    return ne.evaluate('source_E * exp(const * product)')


def get_model():
    return np.sum(get_model_A() * get_model_B(), axis=-1)


###############
# COMPUTATIONS
###############
E = get_model()
I = np.abs(E)**2
D = integ(I)


##########
# DISPLAY
##########
mp.figure()
mp.imshow(np.log10(I), interpolation='nearest', origin='lower')
mp.autoscale(False)
qubic.detector.plot(transform=detector_plane.topixel)
mp.figure()
mp.imshow(D, interpolation='nearest')
mp.gca().format_coord = lambda x, y: 'x={} y={} z={}'.format(x, y, D[x, y])
mp.show()
print('Given {} horns, we get {} W in the detector plane and {} W in the detec'
      'tors.'.format(nhorn_open, np.sum(I), np.sum(D)))



# coding: utf-8
from __future__ import division

import numpy as np
import os
import re
from astropy.io import fits
from glob import glob
from os.path import join
from pysimulators import Layout, LayoutGridCircles

__all__ = ['QubicCalibration']

PATH = join(os.path.dirname(__file__), 'calfiles')

FILE_DETARRAY = 'CalQubic_DetArray_v*.fits'
FILE_HORNARRAY = 'CalQubic_HornArray_v*.fits'
FILE_OPTICS = 'CalQubic_Optics_v*.fits'
FILE_PRIMBEAM = 'CalQubic_PrimBeam_v*.fits'


class QubicCalibration(object):
    """
    Class representing the QUBIC calibration tree. It stores the calibration
    file names and "hardcoded" values and provides access to them.

    If the path name of a calibration file is relative, it is first searched
    relatively to the working directory and if not found, in the calibration
    path.

    """
    def __init__(self, path=PATH, detarray=FILE_DETARRAY,
                 hornarray=FILE_HORNARRAY, optics=FILE_OPTICS,
                 primbeam=FILE_PRIMBEAM):
        """
        Parameters
        ----------
        path : str, optional
            The directory path of the calibration tree. The default one is
            the one that is contained in the qubic package.
        detarray : str, optional
            The detector array layout calibration file name.
        hornarray : str, optional
            The horn array layout calibration file name.
        optics : str, optional
            The optics parameters calibration file name.
        primbeam : str, optional
            The primary beam parameter calibration file name.

        """
        self.path = os.path.abspath(path)
        self.detarray = self._newest(detarray)
        self.hornarray = self._newest(hornarray)
        self.optics = self._newest(optics)
        self.primbeam = self._newest(primbeam)

    def __str__(self):
        state = [('path', self.path),
                 ('detarray', self.detarray),
                 ('hornarray', self.hornarray),
                 ('optics', self.optics),
                 ('primbeam', self.primbeam)]
        return '\n'.join([a + ': ' + repr(v) for a, v in state])

    __repr__ = __str__

    def get(self, name, *args):
        """
        Access calibration files.

        Parameters
        ----------
        name : str
            One of the following:
                - 'detarray'
                - 'hornarray'
                - 'optics'
                - 'primbeam'

        """
        if name == 'detarray':
            hdus = fits.open(self.detarray)
            version = hdus[0].header['format version']
            corner = hdus[2].data
            shape = corner.shape[:-2]
            n = shape[0] * shape[1]
            if version == '1.0':
                removed = np.zeros(shape, bool)
                index = np.arange(n, dtype=np.int16).reshape(shape)
                quadrant = np.zeros(shape, np.int8)
            else:
                removed = hdus[3].data.view(bool)
                index = hdus[4].data
                quadrant = hdus[5].data
            return shape, corner, removed, index, quadrant

        elif name == 'hornarray':
            hdus = fits.open(self.hornarray)
            version = hdus[0].header['format version']
            if version == '1.0':
                h = hdus[0].header
                spacing = h['spacing']
                center = hdus[1].data
                shape = center.shape[:-1]
                layout = Layout(shape, center=center, radius=h['innerrad'])
                layout.spacing = spacing
            elif version == '2.0':
                h = hdus[0].header
                spacing = h['spacing']
                xreflection = h['xreflection']
                yreflection = h['yreflection']
                radius = h['radius']
                selection = ~hdus[1].data.view(bool)
                layout = LayoutGridCircles(
                    removed.shape, spacing, selection=selection, radius=radius,
                    xreflection=xreflection, yreflection=yreflection)
            else:
                h = hdus[1].header
                spacing = h['spacing']
                xreflection = h['xreflection']
                yreflection = h['yreflection']
                angle = h['angle']
                radius = h['radius']
                selection = ~hdus[2].data.view(bool)
                shape = selection.shape
                layout = LayoutGridCircles(
                    shape, spacing, selection=selection, radius=radius,
                    xreflection=xreflection, yreflection=yreflection,
                    angle=angle, startswith1=True, id=None)
                layout.id = np.arange(len(layout))
            layout.kappa = h['kappa']
            return layout

        elif name == 'optics':
            header = fits.open(self.optics)[0].header
            return {'focal length': header['flength']}

        elif name == 'primbeam':
            header = fits.open(self.primbeam)[0].header
            fwhm0_deg = header['fwhm']
            return fwhm0_deg
            # nu0 = hdu['freq']
            # return lambda nu: fwhm0_deg * (nu0 / nu)

        raise ValueError("Invalid calibration item: '{}'".format(name))

    def _newest(self, filename):
        if '*' not in filename:
            if not os.path.exists(filename):
                filename = join(self.path, filename)
            if not os.path.exists(filename):
                raise ValueError("No calibration file '{}'.".format(filename))
            return os.path.abspath(filename)

        filenames = glob(filename)
        if len(filenames) == 0:
            filename = join(self.path, filename)
            filenames = glob(filename)
            if len(filenames) == 0:
                raise ValueError("No calibration files '{}'.".format(filename))
        regex = re.compile(filename.replace('*', '(.*)'))
        version = sorted(regex.search(f).group(1) for f in filenames)[-1]
        return os.path.abspath(filename.replace('*', version))



# coding: utf-8
from __future__ import division

import astropy.io.fits as pyfits
import healpy as hp
import numpy as np
import operator
import os
import time
import yaml
from glob import glob
from pyoperators import (
    BlockColumnOperator, BlockDiagonalOperator, BlockRowOperator,
    MPIDistributionIdentityOperator, I, MPI, proxy_group, rule_manager)
from pyoperators.utils import ifirst
from pyoperators.utils.mpi import as_mpi
from pysimulators import Acquisition, ProjectionOperator
from pysimulators.interfaces.healpy import HealpixConvolutionGaussianOperator
from .calibration import QubicCalibration
from .instrument import QubicInstrument, SimpleInstrument
from .scene import QubicScene

__all__ = ['QubicAcquisition',
           'SimpleAcquisition']


class SimpleAcquisition(Acquisition):
    """
    The Acquisition class for the single-peak simple instrument

    """
    def __init__(self, instrument, sampling, scene=None, block=None,
                 calibration=None, detector_sigma=10, detector_fknee=0,
                 detector_fslope=1, detector_ncorr=10, detector_ngrids=None,
                 detector_tau=0.01, synthbeam_dtype=np.float32,
                 kind='IQU', nside=256, max_nbytes=None,
                 nprocs_instrument=None, nprocs_sampling=None, comm=None):
        """
        acq = SimpleAcquisition(instrument, sampling, [scene=|kind=, nside=],
                                nprocs_instrument=, nprocs_sampling=, comm=)
        acq = SimpleAcquisition(band, sampling, [scene=|kind=, nside=],
                                nprocs_instrument=, nprocs_sampling=, comm=)

        Parameters
        ----------
        band : int
            The sky frequency, in GHz.
        scene : QubicScene, optional
            The discretized observed scene (the sky).
        block : tuple of slices, optional
            Partition of the samplings.
        kind : 'I', 'QU', 'IQU', optional
            The sky kind: 'I' for intensity-only, 'QU' for Q and U maps,
            and 'IQU' for intensity plus QU maps.
        nside : int, optional
            The Healpix scene's nside.
        instrument : SimpleInstrument, optional
            Module name (only 'monochromatic' for now), or a QubicInstrument
            instance.
        calibration : QubicCalibration, optional
            The calibration tree.
        detector_sigma : array-like, optional
            The standard deviation of the detector white noise component.
        detector_fknee : array-like, optional
            The detector 1/f knee frequency in Hertz.
        detector_fslope : array-like, optional
            The detector 1/f slope index.
        detector_ncorr : int, optional
            The detector 1/f correlation length.
        detector_ngrids : int, optional
            Number of detector grids.
        detector_tau : array-like, optional
            The detector time constants in seconds.
        synthbeam_dtype : dtype, optional
            The data type for the synthetic beams (default: float32).
            It is the dtype used to store the values of the pointing matrix.
        max_nbytes : int or None, optional
            Maximum number of bytes to be allocated for the acquisition's
            operator.
        nprocs_instrument : int
            For a given sampling slice, number of procs dedicated to
            the instrument.
        nprocs_sampling : int
            For a given detector slice, number of procs dedicated to
            the sampling.
        comm : mpi4py.MPI.Comm
            The acquisition's MPI communicator. Note that it is transformed
            into a 2d cartesian communicator before being stored as the 'comm'
            attribute. The following relationship must hold:
                comm.size = nprocs_instrument * nprocs_sampling

        """
        if not isinstance(instrument, SimpleInstrument):
            if scene is not None:
                raise TypeError('Invalid calling sequence.')
            scene = QubicScene(instrument, kind=kind, nside=nside)
            if detector_ngrids is None:
                detector_ngrids = 1 if scene.kind == 'I' else 2
            instrument = SimpleInstrument(
                calibration=calibration, detector_fknee=detector_fknee,
                detector_fslope=detector_fslope, detector_ncorr=detector_ncorr,
                detector_ngrids=detector_ngrids, detector_sigma=detector_sigma,
                detector_tau=detector_tau, synthbeam_dtype=synthbeam_dtype)
        elif scene is None:
            scene = QubicScene(150, kind=kind, nside=nside)
        Acquisition.__init__(
            self, instrument, sampling, scene, block,
            max_nbytes=max_nbytes, nprocs_instrument=nprocs_instrument,
            nprocs_sampling=nprocs_sampling, comm=comm)

    def get_coverage(self, out=None):
        """
        Return the acquisition scene coverage given by P.T(1).

        """
        P = self.get_projection_operator()
        if isinstance(P, ProjectionOperator):
            out = P.pT1(out=out)
        else:
            out = P.operands[0].pT1(out=out)
            for op in P.operands[1:]:
                op.pT1(out, operation=operator.iadd)
        self.get_distribution_operator().T(out, out=out)
        return out

    def get_hitmap(self, nside=None):
        """
        Return a healpy map whose values are the number of times a pointing
        hits the pixel.

        """
        if nside is None:
            nside = self.scene.nside
        ipixel = self.sampling.healpix(nside)
        npixel = 12 * nside**2
        hit = np.histogram(ipixel, bins=npixel, range=(0, npixel))[0]
        self.sampling.comm.Allreduce(MPI.IN_PLACE, as_mpi(hit), op=MPI.SUM)
        return hit

    def get_convolution_peak_operator(self, fwhm=np.radians(0.64883707),
                                      **keywords):
        """
        Return an operator that convolves the Healpix sky by the gaussian
        kernel that, if used in conjonction with the peak sampling operator,
        best approximates the synthetic beam.

        Parameters
        ----------
        fwhm : float, optional
            The Full Width Half Maximum of the gaussian.

        """
        return HealpixConvolutionGaussianOperator(fwhm=fwhm, **keywords)

    def get_detector_response_operator(self):
        """
        Return the operator for the bolometer responses.

        """
        return BlockDiagonalOperator(
            [self.instrument.get_detector_response_operator(self.sampling[b])
             for b in self.block], axisin=1)

    def get_distribution_operator(self):
        """
        Return the MPI distribution operator.

        """
        return MPIDistributionIdentityOperator(self.comm)

    def get_hwp_operator(self):
        """
        Return the operator for the bolometer responses.

        """
        return BlockDiagonalOperator(
            [self.instrument.get_hwp_operator(self.sampling[b], self.scene)
             for b in self.block], axisin=1)

    def get_operator(self):
        """
        Return the operator of the acquisition.

        """
        projection = self.get_projection_operator()
        hwp = self.get_hwp_operator()
        polarizer = self.get_polarizer_operator()
        response = self.get_detector_response_operator()
        distribution = self.get_distribution_operator()

        with rule_manager(inplace=True):
            H = response * polarizer * (hwp * projection) * distribution
        if self.scene == 'QU':
            H = self.get_subtract_grid_operator() * H
        return H

    def get_operator_nbytes(self):
        # return self.get_invntt_nbytes() + self.get_projection_nbytes()
        return self.get_projection_nbytes()

    def get_polarizer_operator(self):
        """
        Return operator for the polarizer grid.

        """
        return BlockDiagonalOperator(
            [self.instrument.get_polarizer_operator(
                self.sampling[b], self.scene) for b in self.block], axisin=1)

    def get_projection_operator(self, verbose=True):
        """
        Return the projection operator for the peak sampling.

        Parameters
        ----------
        verbose : bool, optional
            If true, display information about the memory allocation.

        """
        f = self.instrument.get_projection_operator
        if len(self.block) == 1:
            return BlockColumnOperator(
                [f(self.sampling[b], self.scene, verbose=verbose)
                 for b in self.block], axisout=1)
        #XXX HACK
        def callback(i):
            p = f(self.sampling[self.block[i]], self.scene, verbose=False)
            return p
        shapeouts = [(len(self.instrument), s.stop-s.start) +
                      self.scene.shape[1:] for s in self.block]
        proxies = proxy_group(len(self.block), callback, shapeouts=shapeouts)
        return BlockColumnOperator(proxies, axisout=1)

    def get_add_grids_operator(self):
        """ Return operator to add signal from detector pairs. """
        nd = len(self.instrument)
        if nd % 2 != 0:
            raise ValueError('Odd number of detectors.')
        partitionin = 2 * (len(self.instrument) // 2,)
        return BlockRowOperator([I, I], axisin=0, partitionin=partitionin)

    def get_subtract_grids_operator(self):
        """ Return operator to subtract signal from detector pairs. """
        nd = len(self.instrument)
        if nd % 2 != 0:
            raise ValueError('Odd number of detectors.')
        partitionin = 2 * (len(self.instrument) // 2,)
        return BlockRowOperator([I, -I], axisin=0, partitionin=partitionin)

    @classmethod
    def load(cls, filename, instrument=None, selection=None):
        """
        Load a QUBIC acquisition, and info.

        obs, info = QubicAcquisition.load(filename, [instrument=None,
                                          selection=None])

        Parameters
        ----------
        filename : string
           The QUBIC acquisition file name.
        instrument : QubicInstrument, optional
           The Qubic instrumental setup.
        selection : integer or sequence of
           The indices of the pointing blocks to be selected to construct
           the pointing configuration.

        Returns
        -------
        obs : QubicAcquisition
            The QUBIC acquisition instance as read from the file.
        info : string
            The info file stored alongside the acquisition.

        """
        if not isinstance(filename, str):
            raise TypeError("The input filename is not a string.")
        if instrument is None:
            instrument = cls._get_instrument_from_file(filename)
        with open(os.path.join(filename, 'info.txt')) as f:
            info = f.read()
        ptg, ptg_id = cls._get_files_from_selection(filename, 'ptg', selection)
        return QubicAcquisition(instrument, ptg, selection=selection,
                                block_id=ptg_id), info

    @classmethod
    def _load_observation(cls, filename, instrument=None, selection=None):
        """
        Load a QUBIC acquisition, info and TOD.

        obs, tod, info = QubicAcquisition._load_observation(filename,
                             [instrument=None, selection=None])

        Parameters
        ----------
        filename : string
           The QUBIC acquisition file name.
        instrument : QubicInstrument, optional
           The Qubic instrumental setup.
        selection : integer or sequence of
           The indices of the pointing blocks to be selected to construct
           the pointing configuration.

        Returns
        -------
        obs : QubicAcquisition
           The QUBIC acquisition instance as read from the file.
        tod : ndarray
           The stored time-ordered data.
        info : string
           The info file stored alongside the acquisition.

        """
        obs, info = cls.load(filename, instrument=instrument,
                             selection=selection)
        tod, tod_id = cls._get_files_from_selection(filename, 'tod', selection)
        if len(tod) != len(obs.block):
            raise ValueError('The number of pointing and tod files is not the '
                             'same.')
        if any(p != t for p, t in zip(obs.block.identifier, tod_id)):
            raise ValueError('Incompatible pointing and tod files.')
        tod = np.hstack(tod)
        return obs, tod, info

    @classmethod
    def load_simulation(cls, filename, instrument=None, selection=None):
        """
        Load a simulation, including the QUBIC acquisition, info, TOD and
        input map.

        obs, input_map, tod, info = QubicAcquisition.load_simulation(
            filename, [instrument=None, selection=None])

        Parameters
        ----------
        filename : string
           The QUBIC acquisition file name.
        instrument : QubicInstrument, optional
           The Qubic instrumental setup.
        selection : integer or sequence of
           The indices of the pointing blocks to be selected to construct
           the pointing configuration.

        Returns
        -------
        obs : QubicAcquisition
           The QUBIC acquisition instance as read from the file.
        input_map : Healpy map
           The simulation input map.
        tod : Tod
           The stored time-ordered data.
        info : string
           The info file of the simulation.

        """
        obs, tod, info = cls._load_observation(filename, instrument=instrument,
                                               selection=selection)
        input_map = hp.read_map(os.path.join(filename, 'input_map.fits'))
        return obs, input_map, tod, info

    def save(self, filename, info):
        """
        Write a Qubic acquisition to disk.

        Parameters
        ----------
        filename : string
            The output path of the directory in which the acquisition will be
            saved.
        info : string
            All information deemed necessary to describe the acquisition.

        """
        self._save_acquisition(filename, info)
        self._save_ptg(filename)

    def _save_observation(self, filename, tod, info):
        """
        Write a QUBIC acquisition to disk with a TOD.

        Parameters
        ----------
        filename : string
            The output path of the directory in which the simulation will be
            saved.
        tod : array-like
            The simulated time ordered data, of shape (ndetectors, npointings).
        info : string
            All information deemed necessary to describe the simulation.

        """
        self._save_acquisition(filename, info)
        self._save_ptg_tod(filename, tod)

    def save_simulation(self, filename, input_map, tod, info):
        """
        Write a QUBIC acquisition to disk with a TOD and an input image.

        Parameters
        ----------
        filename : string
            The output path of the directory in which the simulation will be
            saved.
        input_map : ndarray, optional
            For simulations, the input Healpix map.
        tod : array-like
            The simulated time ordered data, of shape (ndetectors, npointings).
        info : string
            All information deemed necessary to describe the simulation.

        """
        self._save_observation(filename, tod, info)
        hp.write_map(os.path.join(filename, 'input_map.fits'), input_map)

    def _save_acquisition(self, filename, info):
        # create directory
        try:
            os.mkdir(filename)
        except OSError:
            raise OSError("The path '{}' already exists.".format(filename))

        # instrument state
        with open(os.path.join(filename, 'instrument.txt'), 'w') as f:
            f.write(str(self.instrument))

        # info file
        with open(os.path.join(filename, 'info.txt'), 'w') as f:
            f.write(info)

    def _save_ptg(self, filename):
        for b in self.block:
            postfix = self._get_time_id() + '.fits'
            ptg = self.pointing[b.start:b.stop]
            file_ptg = os.path.join(filename, 'ptg_' + postfix)
            hdu_ptg = pyfits.PrimaryHDU(ptg)
            pyfits.HDUList([hdu_ptg]).writeto(file_ptg)

    def _save_ptg_tod(self, filename, tod):
        for b in self.block:
            postfix = self._get_time_id() + '.fits'
            p = self.pointing[b.start:b.stop]
            t = tod[:, b.start:b.stop]
            file_ptg = os.path.join(filename, 'ptg_' + postfix)
            file_tod = os.path.join(filename, 'tod_' + postfix)
            hdu_ptg = pyfits.PrimaryHDU(p)
            hdu_tod = pyfits.PrimaryHDU(t)
            pyfits.HDUList([hdu_ptg]).writeto(file_ptg)
            pyfits.HDUList([hdu_tod]).writeto(file_tod)

    @staticmethod
    def _get_instrument_from_file(filename):
        with open(os.path.join(filename, 'instrument.txt')) as f:
            lines = f.readlines()[1:]
        sep = ifirst(lines, '\n')
        keywords_instrument = yaml.load(''.join(lines[:sep]))
        name = keywords_instrument.pop('name')
        keywords_calibration = yaml.load(''.join(lines[sep+2:]))
        calibration = QubicCalibration(**keywords_calibration)
        return QubicInstrument(name, calibration, **keywords_instrument)

    @staticmethod
    def _get_files_from_selection(filename, filetype, selection):
        """ Read files from selection, without reading them twice. """
        files = sorted(glob(os.path.join(filename, filetype + '*.fits')))
        if selection is None:
            return [pyfits.open(f)[0].data for f in files], \
                   [f[-13:-5] for f in files]
        if not isinstance(selection, (list, tuple)):
            selection = (selection,)
        iuniq, inv = np.unique(selection, return_inverse=True)
        uniq_data = [pyfits.open(files[i])[0].data for i in iuniq]
        uniq_id = [files[i][-13:-5] for i in iuniq]
        return [uniq_data[i] for i in inv], [uniq_id[i] for i in inv]

    @staticmethod
    def _get_time_id():
        t = time.time()
        return time.strftime('%Y-%m-%d_%H:%M:%S',
                             time.localtime(t)) + '{:.9f}'.format(t-int(t))[1:]


class QubicAcquisition(SimpleAcquisition):
    """
    The QubicAcquisition class, which combines the instrument, sampling and
    scene models.

    """
    def __init__(self, instrument, sampling, scene=None, block=None,
                 calibration=None, detector_sigma=10, detector_fknee=0,
                 detector_fslope=1, detector_ncorr=10, detector_ngrids=None,
                 detector_tau=0.01, primary_beam=None, secondary_beam=None,
                 synthbeam_dtype=np.float32, synthbeam_fraction=0.99,
                 kind='IQU', nside=256, max_nbytes=None,
                 nprocs_instrument=None, nprocs_sampling=None, comm=None):
        """
        acq = QubicAcquisition(instrument, sampling, [scene=|kind=, nside=],
                               nprocs_instrument=, nprocs_sampling=, comm=)
        acq = QubicAcquisition(band, sampling, [scene=|kind=, nside=],
                               nprocs_instrument=, nprocs_sampling=, comm=)

        Parameters
        ----------
        band : int
            The sky frequency, in GHz.
        scene : QubicScene, optional
            The discretized observed scene (the sky).
        block : tuple of slices, optional
            Partition of the samplings.
        kind : 'I', 'QU', 'IQU', optional
            The sky kind: 'I' for intensity-only, 'QU' for Q and U maps,
            and 'IQU' for intensity plus QU maps.
        nside : int, optional
            The Healpix scene's nside.
        instrument : QubicInstrument, optional
            Module name (only 'monochromatic' for now), or a QubicInstrument
            instance.
        calibration : QubicCalibration, optional
            The calibration tree.
        detector_sigma : array-like, optional
            The standard deviation of the detector white noise component.
        detector_fknee : array-like, optional
            The detector 1/f knee frequency in Hertz.
        detector_fslope : array-like, optional
            The detector 1/f slope index.
        detector_ncorr : int, optional
            The detector 1/f correlation length.
        detector_ngrids : int, optional
            Number of detector grids.
        detector_tau : array-like, optional
            The detector time constants in seconds.
        primary_beam : function f(theta [rad], phi [rad])
            The primary beam transmission function.
        secondary_beam : function f(theta [rad], phi [rad])
            The secondary beam transmission function.
        synthbeam_dtype : dtype, optional
            The data type for the synthetic beams (default: float32).
            It is the dtype used to store the values of the pointing matrix.
        synthbeam_fraction: float, optional
            The fraction of significant peaks retained for the computation
            of the synthetic beam.
        max_nbytes : int or None, optional
            Maximum number of bytes to be allocated for the acquisition's
            operator.
        nprocs_instrument : int
            For a given sampling slice, number of procs dedicated to
            the instrument.
        nprocs_sampling : int
            For a given detector slice, number of procs dedicated to
            the sampling.
        comm : mpi4py.MPI.Comm
            The acquisition's MPI communicator. Note that it is transformed
            into a 2d cartesian communicator before being stored as the 'comm'
            attribute. The following relationship must hold:
                comm.size = nprocs_instrument * nprocs_sampling

        """
        if not isinstance(instrument, QubicInstrument):
            if scene is not None:
                raise TypeError('Invalid calling sequence.')
            scene = QubicScene(instrument, kind=kind, nside=nside)
            if detector_ngrids is None:
                detector_ngrids = 1 if scene.kind == 'I' else 2
            instrument = QubicInstrument(
                calibration=calibration, detector_fknee=detector_fknee,
                detector_fslope=detector_fslope, detector_ncorr=detector_ncorr,
                detector_ngrids=detector_ngrids, detector_sigma=detector_sigma,
                detector_tau=detector_tau, primary_beam=primary_beam,
                secondary_beam=secondary_beam, synthbeam_dtype=synthbeam_dtype,
                synthbeam_fraction=synthbeam_fraction)
        SimpleAcquisition.__init__(
            self, instrument, sampling, scene, block=block,
            max_nbytes=max_nbytes, nprocs_instrument=nprocs_instrument,
            nprocs_sampling=nprocs_sampling, comm=comm)



#!/usr/bin/env python
import numpy as np
import os
from numpy.distutils.core import setup
from numpy.distutils.extension import Extension
from hooks import get_cmdclass, get_version

# force sdist to copy files
delattr(os, 'link')

VERSION = '4.1'

name = 'qubic'
long_description = open('README.rst').read()
keywords = 'scientific computing'
platforms = 'MacOS X,Linux,Solaris,Unix,Windows'
extra_f90_compile_args = ['-cpp -fopenmp -fpack-derived']

ext_modules = [Extension('qubic._flib',
                         sources=['src/polarization.f90.src'],
                         extra_f90_compile_args=extra_f90_compile_args,
                         include_dirs=['.', np.get_include()],
                         libraries=['gomp'])]

setup(name=name,
      version=get_version(name, VERSION),
      description='Simulation and map-making tools for the QUBIC experiment.',
      long_description=long_description,
      url='',
      author='Pierre Chanial',
      author_email='pierre.chanial@apc.univ-paris7.fr',
      install_requires=['progressbar',
                        'pyoperators>=0.12.12',
                        'pysimulators>=1.0',
                        'healpy>=0.6.1',
                        'pyYAML'],
      packages=['qubic', 'qubic/io'],
      package_data={'qubic': ['calfiles/*fits', 'data/*', 'scripts/*py']},
      platforms=platforms.split(','),
      keywords=keywords.split(','),
      cmdclass=get_cmdclass(),
      ext_modules=ext_modules,
      license='CeCILL-B',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2 :: Only',
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Astronomy'])



from __future__ import division

import numpy as np
import os
import qubic
import shutil
from numpy.testing import assert_equal
from pyoperators.utils.testing import assert_same, skiptest
from qubic import (
    QubicCalibration, QubicAcquisition, QubicInstrument, QubicSampling,
    create_random_pointings, gal2equ, map2tod)
from qubic.io import read_map
from uuid import uuid1

outpath = ''


def setup():
    global outpath
    outpath = 'test-' + str(uuid1())[:8]
    os.mkdir(outpath)


def teardown():
    shutil.rmtree(outpath)

DATAPATH = os.path.join(os.path.dirname(qubic.__file__), 'data')
input_map = read_map(os.path.join(DATAPATH, 'syn256_pol.fits'))
np.random.seed(0)
center = gal2equ(0, 90)
sampling = create_random_pointings(center, 100, 10)

ptg = [1., 0, 180]
ptg2 = [1., 1, 180]
pta = np.asarray(ptg)
pta2 = np.asarray(ptg2)
ptgs = ptg, [ptg], [ptg, ptg], [[ptg, ptg], [ptg2, ptg2, ptg2]], \
       pta, [pta], [pta, pta], [[pta, pta], [pta2, pta2, pta2]], \
       np.asarray([pta]), np.asarray([pta, pta]), [np.asarray([pta, pta])], \
       [np.asarray([pta, pta]), np.asarray([pta2, pta2, pta2])]
block_n = [[1], [1], [2], [2, 3],
           [1], [1], [2], [2, 3],
           [1], [2], [2], [2, 3], [2, 3]]
caltree = QubicCalibration(detarray='CalQubic_DetArray_v1.fits')
removed = np.ones((32, 32), dtype=bool)
removed[30:, 30:] = False
qubic = QubicInstrument(calibration=caltree, nu=160e9,
                        removed=removed, synthbeam_fraction=0.99)


@skiptest
def test_pointing():
    def func(p, n):
        p = np.asarray(p)
        smp = QubicSampling(azimuth=p[..., 0], elevation=p[..., 1],
                            pitch=p[..., 2])
        acq = QubicAcquisition(qubic, smp)
        assert_equal(acq.block.n, n)
        assert_equal(len(acq.pointing), sum(n))
    for p, n in zip(ptgs, block_n):
        yield func, p, n


@skiptest
def test_load_save():
    info = 'test\nconfig'

    def func_acquisition(obs, info):
        filename_ = os.path.join(outpath, 'config-' + str(uuid1()))
        obs.save(filename_, info)
        obs2, info2 = QubicAcquisition.load(filename_)
        assert_equal(str(obs), str(obs2))
        assert_equal(info, info2)

    def func_observation(obs, tod, info):
        filename_ = os.path.join(outpath, 'obs-' + str(uuid1()))
        obs._save_observation(filename_, tod, info)
        obs2, tod2, info2 = QubicAcquisition._load_observation(filename_)
        assert_equal(str(obs), str(obs2))
        assert_equal(tod, tod2)
        assert_equal(info, info2)

    def func_simulation(obs, input_map, tod, info):
        filename_ = os.path.join(outpath, 'simul-' + str(uuid1()))
        obs.save_simulation(filename_, input_map, tod, info)
        obs2, input_map2, tod2, info2 = QubicAcquisition.load_simulation(
            filename_)
        assert_equal(str(obs), str(obs2))
        assert_equal(input_map, input_map2)
        assert_equal(tod, tod2)
        assert_equal(info, info2)

    for p in ptgs:
        p = np.asarray(p)
        smp = QubicSampling(azimuth=p[..., 0], elevation=p[..., 1],
                            pitch=p[..., 2])
        acq = QubicAcquisition(qubic, smp)
        P = acq.get_projection_peak_operator()
        tod = P(input_map)
        yield func_acquisition, acq, info
        yield func_observation, acq, tod, info
        yield func_simulation, acq, input_map, tod, info


def test_add_subtract_grid_operator():
    acq = QubicAcquisition(150, sampling)
    tod = map2tod(acq, input_map, convolution=False)
    add = acq.get_add_grids_operator()
    sub = acq.get_subtract_grids_operator()
    assert_same(add(tod), tod[:992] + tod[992:])
    assert_same(sub(tod), tod[:992] - tod[992:])



from __future__ import division

import healpy as hp
import numpy as np
from astropy.io import fits
from collections import Container

__all__ = ['read_map', 'write_map']

_default_extnames = {
    1: ('I_STOKES',),
    3: ('I_STOKES', 'Q_STOKES', 'U_STOKES'),
    6: ('II', 'IQ', 'IU', 'QQ', 'QU', 'UU')}


def read_map(filename, field=None, dtype=float, nest=False, partial=False,
             h=False):
    """
    Read one or more compressed Healpix maps from a FITS file.

    Parameters
    ----------
    filename : str
        The FITS file name.
    field : int, str or tuple of int or str, optional
        The fields to be read. Default: all. By convention 0 is temperature,
        1 is Q, 2 is U. It can be a tuple to read multiple columns (0, 1, 2).
        Extension names can also be used: 'I_STOKES'
    dtype : data type, optional
        Force the conversion to some data type. Default: np.float64
    nest : bool, optional
        If True, return the map in NEST ordering, otherwise in RING ordering;
        use fits keyword ORDERING to decide whether conversion is needed or not
        If None, no conversion is performed.
    partial : bool, optional
        If True, return the partial map and the mask of valid pixels.
    h : bool, optional
        If True, return also the header. Default: False.

    Returns
    -------
    maps [, mask[, header]] : array of shape (N, M)
        The Healpix maps, where M is the number of maps and N the number
        of pixels.
    mask : bool array
        The selection mask (True means valid) if the partial keyword is
        set to True.
    header : astropy.io.fits.Header
        The optional FITS header if the keyword h has been set to True.

    """
    if isinstance(filename, file):
        modes = {'+ab': 'update'}
        mode = modes.get(''.join(sorted(filename.mode)), 'readonly')
    else:
        mode = 'readonly'
    hdus = fits.open(filename, mode=mode)
    header0 = hdus[0].header
    if 'format' not in header0 or header0['format'] != 'HPX_QB':
        # fall back to healpy's read_map
        if len(hdus) != 2 or 'pixtype' not in hdus[1].header or \
           hdus[1].header['pixtype'] != 'HEALPIX':
            raise TypeError('This file cannot be read as a Healpix map.')
        header = hdus[1].header
        if field is None:
            field = range(header['TFIELDS'])
        out = hp.read_map(filename, dtype=dtype, nest=nest, field=field, h=h,
                          verbose=False)
        if h:
            out, header = np.column_stack(out[:-1]), fits.Header(out[-1])
            if partial:
                return out, None, header
            return out, header
        out = np.column_stack(out)
        if partial:
            return out, None
        return out

    nside = header0['nside']
    npix = 12 * nside**2
    ordering = header0['ordering']
    if field is None:
        field = range(header0['nmaps'])
    elif not isinstance(field, Container) or isinstance(field, str):
        field = (field,)

    npix_ = npix
    if header0['hasmask'] and header0['hasmask']:
        mask = hdus[1].data.view(bool)
        mask_ = mask
        ifirst = 2
        outfunc = lambda shape, **keywords: np.full(shape, np.nan, **keywords)
        if partial:
            npix_ = np.sum(mask)
            mask_ = Ellipsis
    else:
        mask = None
        mask_ = Ellipsis
        ifirst = 1
        outfunc = np.empty
    if dtype is None:
        bitpix_table = {8: np.uint8, 16: np.int16, 32: np.int32, 64: np.int64,
                        -32: np.float32, -64: np.float64}
        dtype = bitpix_table[hdus[ifirst].header['bitpix']]
    out = outfunc((len(field), npix_), dtype=dtype, order='f')

    if partial and (nest and ordering == 'RING' or
                    not nest and ordering == 'NESTED'):
        raise NotImplementedError('The partial keyword cannot be set while cha'
                                  'nging the ordering.')
    reorder = False
    if nest and ordering == 'RING':
        indexing = hp.pixelfunc.nest2ring(nside, np.arange(npix))
        reorder = True
    elif not nest and ordering == 'NESTED':
        indexing = hp.pixelfunc.ring2nest(nside, np.arange(npix))
        reorder = True

    for i, field_ in enumerate(field):
        ihdu = field_ if isinstance(field_, str) else field_ + ifirst
        out_ = out[i]
        out_[mask_] = hdus[ihdu].data
        if reorder:
            np.take(out_, indexing, out=out_)
        try:
            out_[hp.pixelfunc.mask_bad(out_)] = hp.UNSEEN
        except OverflowError:
            pass

    if len(field) == 1:
        out = out[0]
    else:
        out = out.T
    if partial or h:
        out = (out,)
    if partial:
        out += (mask,)
    if h:
        out += (header0,)
    return out


def write_map(filename, map, mask=None, nest=False, dtype=np.float32,
              coord=None, extnames=None, compress=True):
    """
    Write one or more compressed (complete or partial) Healpix maps
    as FITS file.

    Parameters
    ----------
    filename : str
        The FITS file name.
    map : array of shape (N) or (N, M)
        The partial or complete input Healpix maps.
    mask : bool array, optional
        The mask controlling partial maps, such as complete[mask] == partial
        (True means valid).
    nest : bool, optional
        If True, ordering scheme is assumed to be NESTED, otherwise, RING.
        Default: RING. The map ordering is not modified by this function,
        the input map array should already be in the desired ordering.
    coord : str
        The coordinate system, typically 'E' for Ecliptic, 'G' for Galactic or
        'C' for Celestial (equatorial).
    extnames : str or list
        The FITS extension names, by default, we use:
            - I_STOKES for 1 component,
            - I/Q/U_STOKES for 3 components,
            - II, IQ, IU, QQ, QU, UU for 6 components,
            - DATA_0, DATA_1... otherwise
    """
    map = np.asarray(map, order='f') #XXX astropy issue #2150
    if map.ndim not in (1, 2):
        raise ValueError('Invalid dimensions of the healpix map(s).')
    if map.ndim == 1:
        map = map.reshape(-1, 1)
    nmaps = map.shape[1]
    if mask is not None:
        mask = np.asarray(mask, np.uint8)
        if mask.ndim != 1:
            raise ValueError('Invalid dimensions of healpix the mask.')
        npix = mask.size
    else:
        npix = map.shape[0]
    ordering = 'NESTED' if nest else 'RING'
    nside = hp.npix2nside(npix)

    if compress and map.dtype != int: #XXX avoid crash: astropy issue #2153
        _imageHDU = fits.CompImageHDU
    else:
        _imageHDU = fits.ImageHDU

    primary = fits.PrimaryHDU()
    primary.header['nside'] = nside, 'Resolution parameter of HEALPIX'
    primary.header['ordering'] = (ordering, 'Pixel ordering scheme, '
                                  'either RING or NESTED')
    if coord:
        primary.header['coordsys'] = (coord, 'Ecliptic, Galactic or Celestial '
                                      '(equatorial)')
    primary.header['format'] = 'HPX_QB'
    primary.header['nmaps'] = nmaps
    primary.header['hasmask'] = mask is not None
    hdus = [primary]

    if mask is not None:
        hdu = fits.CompImageHDU(mask)
        hdu.update_ext_name('VALIDMASK')
        hdus.append(hdu)

    if extnames is None:
        extnames = _default_extnames.get(
            nmaps, ('DATA_{}'.format(i + 1) for i in range(nmaps)))

    for m, extname in zip(map.T, extnames):
        hdu = _imageHDU(np.array(m, dtype=dtype, copy=False))
        hdu.update_ext_name(extname)
        hdus.append(hdu)

    fits.HDUList(hdus).writeto(filename, clobber=True)



from __future__ import division
import numpy as np
from pyoperators.utils.testing import assert_same
from qubic import (
    QubicAcquisition, QubicInstrument, QubicScene, create_random_pointings)
from qubic.mapmaking import tod2map_all, tod2map_each


def test():
    instrument = QubicInstrument()[:10]
    sampling = create_random_pointings([0, 90], 30, 5)
    sampling.angle_hwp = np.random.random_integers(0, 7, len(sampling)) * 11.25
    nside = 64
    scenes = QubicScene(150, nside, kind='I'), QubicScene(150, nside)

    def func(scene, max_nbytes, ref1, ref2, ref3, ref4, ref5, ref6):
        acq = QubicAcquisition(instrument, sampling, scene,
                               max_nbytes=max_nbytes)
        sky = np.ones(scene.shape)
        H = acq.get_operator()
        actual1 = H(sky)
        assert_same(actual1, ref1, atol=10)
        actual2 = H.T(actual1)
        assert_same(actual2, ref2, atol=10)
        actual2 = (H.T * H)(sky)
        assert_same(actual2, ref2, atol=10)
        actual3, actual4 = tod2map_all(acq, ref1, disp=False)
        assert_same(actual3, ref3, atol=10000)
        assert_same(actual4, ref4)
        actual5, actual6 = tod2map_each(acq, ref1, disp=False)
        assert_same(actual5, ref5, atol=1000)
        assert_same(actual6, ref6)

    for scene in scenes:
        acq = QubicAcquisition(instrument, sampling, scene)
        sky = np.ones(scene.shape)
        nbytes_per_sampling = acq.get_operator_nbytes() // len(acq.sampling)
        H = acq.get_operator()
        ref1 = H(sky)
        ref2 = H.T(ref1)
        ref3, ref4 = tod2map_all(acq, ref1, disp=False)
        ref5, ref6 = tod2map_each(acq, ref1, disp=False)
        for max_sampling in 10, 29, 30:
            max_nbytes = None if max_sampling is None \
                              else max_sampling * nbytes_per_sampling
            yield (func, scene, max_nbytes, ref1, ref2, ref3, ref4, ref5, ref6)
